results_leukemia = dict(
    has_match=[
        True,
        False,
        True,
        True,
    ],
    detailled_status="PRESENT",
    assign=None,
    texts=[
        "Sydrome myéloprolifératif",
        "Sydrome myéloprolifératif bénin",
        "Patient atteint d'une LAM",
        "Une maladie de Vaquez",
    ],
)



from typing import List, Optional, Set, Union

from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanGetterArg, get_spans
from edsnlp.pipes.qualifiers.base import RuleBasedQualifier
from edsnlp.pipes.terminations import termination as default_termination
from edsnlp.utils.filter import consume_spans, filter_spans
from edsnlp.utils.inclusion import check_inclusion
from edsnlp.utils.resources import get_verbs

from . import patterns


def hypothesis_getter(token: Union[Token, Span]) -> Optional[str]:
    if token._.hypothesis is True:
        return "HYP"
    elif token._.hypothesis is False:
        return "CERT"
    else:
        return None


class HypothesisQualifier(RuleBasedQualifier):
    """
    The `eds.hypothesis` pipeline uses a simple rule-based algorithm to detect spans
    that are speculations rather than certain statements.

    The component looks for five kinds of expressions in the text :

    - preceding hypothesis, ie cues that precede a hypothetical expression
    - following hypothesis, ie cues that follow a hypothetical expression
    - pseudo hypothesis : contain a hypothesis cue, but are not hypothesis
      (eg "pas de doute"/"no doubt")
    - hypothetical verbs : verbs indicating hypothesis (eg "douter")
    - classic verbs conjugated to the conditional, thus indicating hypothesis

    Examples
    --------
    The following snippet matches a simple terminology, and checks whether the extracted
    entities are part of a speculation. It is complete and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    # Dummy matcher
    nlp.add_pipe(eds.matcher(terms=dict(douleur="douleur", fracture="fracture")))
    nlp.add_pipe(eds.hypothesis())

    text = (
        "Le patient est admis le 23 août 2021 pour une douleur au bras. "
        "Possible fracture du radius."
    )

    doc = nlp(text)

    doc.ents
    # Out: (douleur, fracture)

    doc.ents[0]._.hypothesis
    # Out: False

    doc.ents[1]._.hypothesis
    # Out: True
    ```

    Extensions
    ----------
    The `eds.hypothesis` component declares two extensions, on both `Span` and `Token`
    objects :

    1. The `hypothesis` attribute is a boolean, set to `True` if the component predicts
       that the span/token is a speculation.
    2. The `hypothesis_` property is a human-readable string, computed from the
       `hypothesis` attribute. It implements a simple getter function that outputs
       `HYP` or `CERT`, depending on the value of `hypothesis`.

    Performance
    ------------
    The component's performance is measured on three datasets :

    - The ESSAI ([@dalloux2017ESSAI]) and CAS ([@grabar2018CAS]) datasets were developed
      at the CNRS. The two are concatenated.
    - The NegParHyp corpus was specifically developed at APHP's CDW to test the
      component on actual clinical notes, using pseudonymised notes from the APHP's CDW.

    | Dataset   | Hypothesis F1 |
    | --------- | ------------- |
    | CAS/ESSAI | 49%           |
    | NegParHyp | 52%           |

    !!! note "NegParHyp corpus"

        The NegParHyp corpus was built by matching a subset of the MeSH terminology with
        around 300 documents from AP-HP's clinical data warehouse. Matched entities were
        then labelled for negation, speculation and family context.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : Optional[str]
        The component name.
    attr : str
        spaCy's attribute to use
    pseudo : Optional[List[str]]
        List of pseudo hypothesis cues.
    preceding : Optional[List[str]]
        List of preceding hypothesis cues
    following : Optional[List[str]]
        List of following hypothesis cues.
    verbs_hyp : Optional[List[str]]
        List of hypothetical verbs.
    verbs_eds : Optional[List[str]]
        List of mainstream verbs.
    termination : Optional[List[str]]
        List of termination terms.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with the key 'term_attr'
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Deprecated, use `span_getter` instead.

        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    within_ents : bool
        Whether to consider cues within entities.
    explain : bool
        Whether to keep track of cues for each entity.

    Authors and citation
    --------------------
    The `eds.hypothesis` pipeline was developed by AP-HP's Data Science team.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "hypothesis",
        *,
        pseudo: Optional[List[str]] = None,
        preceding: Optional[List[str]] = None,
        following: Optional[List[str]] = None,
        verbs_eds: Optional[List[str]] = None,
        verbs_hyp: Optional[List[str]] = None,
        termination: Optional[List[str]] = None,
        attr: str = "NORM",
        span_getter: Optional[SpanGetterArg] = None,
        on_ents_only: Optional[Union[bool, str, List[str], Set[str]]] = None,
        within_ents: bool = False,
        explain: bool = False,
    ):
        terms = dict(
            pseudo=patterns.pseudo if pseudo is None else pseudo,
            preceding=patterns.preceding if preceding is None else preceding,
            following=patterns.following if following is None else following,
            termination=default_termination if termination is None else termination,
            verbs_eds=patterns.verbs_eds if verbs_eds is None else verbs_eds,
            verbs_hyp=patterns.verbs_hyp if verbs_hyp is None else verbs_hyp,
        )
        terms["verbs_preceding"], terms["verbs_following"] = self.load_verbs(
            verbs_hyp=terms.pop("verbs_hyp"),
            verbs_eds=terms.pop("verbs_eds"),
        )

        super().__init__(
            nlp=nlp,
            name=name,
            attr=attr,
            explain=explain,
            terms=terms,
            on_ents_only=on_ents_only,
            span_getter=span_getter,
        )

        self.within_ents = within_ents
        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        for cls in (Token, Span):
            if not cls.has_extension("hypothesis"):
                cls.set_extension("hypothesis", default=None)

            if not cls.has_extension("hypothesis_"):
                cls.set_extension("hypothesis_", getter=hypothesis_getter)

        if not Span.has_extension("hypothesis_cues"):
            Span.set_extension("hypothesis_cues", default=[])

        if not Doc.has_extension("hypothesis"):
            Doc.set_extension("hypothesis", default=[])

    def load_verbs(
        self,
        verbs_hyp: List[str],
        verbs_eds: List[str],
    ) -> List[str]:
        """
        Conjugate "classic" verbs to conditional, and add hypothesis
        verbs conjugated to all tenses.

        Parameters
        ----------
        verbs_hyp: List of verbs that specifically imply an hypothesis.
        verbs_eds: List of general verbs.

        Returns
        -------
        list of hypothesis verbs conjugated at all tenses and classic
        verbs conjugated to conditional.
        """

        classic_verbs = get_verbs(verbs_eds)
        classic_verbs = classic_verbs.loc[classic_verbs["mode"] == "Conditionnel"]
        list_classic_verbs = list(classic_verbs["term"].unique())

        hypo_verbs = get_verbs(verbs_hyp)
        list_hypo_verbs_preceding = list(hypo_verbs["term"].unique())

        hypo_verbs_following = hypo_verbs.loc[hypo_verbs["tense"] == "Participe Passé"]
        list_hypo_verbs_following = list(hypo_verbs_following["term"].unique())

        return (
            list_hypo_verbs_preceding + list_classic_verbs,
            list_hypo_verbs_following,
        )

    def process(self, doc: Doc) -> Doc:
        matches = self.get_matches(doc)

        terminations = [m for m in matches if m.label_ == "termination"]
        boundaries = self._boundaries(doc, terminations)

        # Removes duplicate matches and pseudo-expressions in one statement
        matches = filter_spans(matches, label_to_remove="pseudo")

        entities = list(get_spans(doc, self.span_getter))
        ents = None

        for start, end in boundaries:
            ents, entities = consume_spans(
                entities,
                filter=lambda s: check_inclusion(s, start, end),
                second_chance=ents,
            )

            sub_matches, matches = consume_spans(
                matches, lambda s: start <= s.start < end
            )

            if self.on_ents_only and not ents:
                continue

            sub_preceding = [m for m in sub_matches if m.label_ == "preceding"]
            sub_following = [m for m in sub_matches if m.label_ == "following"]
            # Verbs preceding negated content
            sub_preceding += [m for m in sub_matches if m.label_ == "verbs_preceding"]
            # Verbs following negated content
            sub_following += [m for m in sub_matches if m.label_ == "verbs_following"]

            if not self.on_ents_only:
                for token in doc[start:end]:
                    token._.hypothesis = any(
                        m.end <= token.i for m in sub_preceding
                    ) or any(m.start > token.i for m in sub_following)

            for ent in ents:
                if self.within_ents:
                    cues = [m for m in sub_preceding if m.end <= ent.end]
                    cues += [m for m in sub_following if m.start >= ent.start]
                else:
                    cues = [m for m in sub_preceding if m.end <= ent.start]
                    cues += [m for m in sub_following if m.start >= ent.end]

                hypothesis = bool(cues)
                ent._.hypothesis = ent._.hypothesis or hypothesis

                if self.explain and hypothesis:
                    ent._.hypothesis_cues += cues

                if not self.on_ents_only and hypothesis:
                    for token in ent:
                        token._.hypothesis = token._.hypothesis or hypothesis

        return doc



TEXT = """
Le patientqsfqfdf bla bla bla
Leucocytes ¦x10*9/L ¦4.97 ¦4.09-11
Hématies ¦x10*12/L¦4.68 ¦4.53-5.79
Hémoglobine ¦g/dL ¦14.8 ¦13.4-16.7
Hématocrite ¦% ¦44.2 ¦39.2-48.6
VGM ¦fL ¦94.4 + ¦79.6-94
TCMH ¦pg ¦31.6 ¦27.3-32.8
CCMH ¦g/dL ¦33.5 ¦32.4-36.3
Plaquettes ¦x10*9/L ¦191 ¦172-398
VMP ¦fL ¦11.5 + ¦7.4-10.8

qdfsdf

2/2Pat : <NOM> <Prenom> |<date> | <ipp> |Intitulé RCP


"""


def test_tables(blank_nlp):
    blank_nlp.add_pipe("eds.normalizer")
    blank_nlp.add_pipe("eds.tables")

    doc = blank_nlp(TEXT)

    assert len(doc.spans["tables"]) == 1

    span = doc.spans["tables"][0]
    df = span._.to_pd_table()
    assert df.iloc[5, 0] == "TCMH "



from edsnlp.utils.examples import find_matches, parse_example, parse_match

example = (
    "Le <ent int_value=1 float_value=0.3>patient</ent> "
    "n'est pas <ent polarity_=NEG negation=true>malade</ent>."
)


def test_find_matches():
    matches = find_matches(example=example)

    assert len(matches) == 2

    match1, match2 = matches

    assert match1.group() == "<ent int_value=1 float_value=0.3>patient</ent>"
    assert match2.group() == "<ent polarity_=NEG negation=true>malade</ent>"


def test_parse_match():
    matches = find_matches(example=example)
    match = parse_match(matches[0])

    assert match.text == "patient"
    assert match.modifiers == "int_value=1 float_value=0.3"


def test_parse_example():
    text, entities = parse_example(example=example)

    assert text == "Le patient n'est pas malade."

    entity1, entity2 = entities

    assert text[entity1.start_char : entity1.end_char] == "patient"
    assert text[entity2.start_char : entity2.end_char] == "malade"

    m1, m2, m3, m4 = entity1.modifiers + entity2.modifiers

    assert m1.key == "int_value"
    assert m1.value == 1

    assert m2.key == "float_value"
    assert m2.value == 0.3

    assert m3.key == "polarity_"
    assert m3.value == "NEG"

    assert m4.key == "negation"
    assert m4.value



default_patterns = dict(
    source="alcohol",
    regex=[
        r"\balco[ol]",
        r"\bethyl",
        r"(?<!(25.?)|(sevrage)).?\boh\b",
        r"exogenose",
        r"delirium.tremens",
    ],
    exclude=[
        dict(
            regex=[
                "occasion",
                "episod",
                "festi",
                "rare",
                "libre",  # OH-libres
                "aigu",
            ],
            window=(-3, 5),
        ),
        dict(
            regex=["pansement", "compress"],
            window=-3,
        ),
    ],
    regex_attr="NORM",
    assign=[
        dict(
            name="stopped",
            regex=r"(?<!non )(?<!pas )(sevr|arret|stop|ancien)",
            window=(-3, 5),
        ),
        dict(
            name="zero_after",
            regex=r"(?=^[a-z]*\s*:?[\s-]*(0|oui|non(?! sevr)))",
            window=6,
        ),
    ],
)



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from edsnlp.core.lazy_collection import LazyCollection

from .converters import get_dict2doc_converter, get_doc2dict_converter


class BaseReader:
    """
    The BaseReader servers as a base class for all readers. It expects two methods:

    - `read_main` method which is called in the main process and should return a
        generator of fragments (like filenames) with their estimated size (number of
        documents)
    - `read_worker` method which is called in the worker processes and receives
        batches of fragments and should return a list of dictionaries (one per
        document), ready to be converted to a Doc object by the converter.

    Additionally, the subclass should define a `DATA_FIELDS` class attribute which
    contains the names of all attributes that should not be copied when the reader is
    copied to the worker processes. This is useful for example when the reader holds a
    reference to a large object like a DataFrame that should not be copied to the
    worker processes.
    """

    DATA_FIELDS = ()

    def read_main(self) -> Iterable[Tuple[Any, int]]:
        raise NotImplementedError()

    def read_worker(self, fragment: Iterable[Any]) -> Iterable[Dict]:
        raise NotImplementedError()

    def worker_copy(self):
        # new reader without data, this will not call __init__ since we use __dict__
        # to set the data
        reader = self.__class__.__new__(self.__class__)
        state = {
            k: v
            for k, v in self.__dict__.items()
            if k not in self.__class__.DATA_FIELDS
        }
        reader.__dict__ = state
        return reader


T = TypeVar("T")


class BaseWriter:
    def write_worker(self, records: Sequence[Any]) -> T:
        raise NotImplementedError()

    def write_main(self, fragments: Iterable[T]):
        raise NotImplementedError()

    def finalize(self):
        return None, 0


class IterableReader(BaseReader):
    DATA_FIELDS = ("data",)

    def __init__(self, data: Iterable):
        self.data = data

        super().__init__()

    def read_main(self) -> Iterable[Tuple[Any, int]]:
        return ((item, 1) for item in self.data)

    def read_worker(self, fragments):
        return [task for task in fragments]


def from_iterable(
    data: Iterable,
    converter: Union[str, Callable] = None,
    **kwargs,
) -> LazyCollection:
    """
    The IterableReader (or `edsnlp.data.from_iterable`) reads a list of Python objects (
    texts, dictionaries, ...) and yields documents by passing them through the
    `converter` if given, or returns them as is.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.from_iterable([{...}], nlp=nlp, converter=...)
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.from_iterable` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.from_iterable([{...}], converter=...)
        ```

    Parameters
    ----------
    data: Iterable
        The data to read
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the JSON rows of the data source to Doc objects
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented
        on the [Converters](/data/converters) page.

    Returns
    -------
    LazyCollection
    """
    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


def to_iterable(
    data: Union[Any, LazyCollection],
    converter: Optional[Union[str, Callable]] = None,
    **kwargs,
):
    """
    `edsnlp.data.to_items` returns an iterator of documents, as converted by the
    `converter`. In comparison to just iterating over a LazyCollection, this will
    also apply the `converter` to the documents, which can lower the data transfer
    overhead when using multiprocessing.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.to_items([doc], converter="omop")
    ```

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects.
    kwargs:
        Additional keyword arguments passed to the converter. These are documented
        on the [Converters](/data/converters) page.
    """
    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data



from edsnlp.core import registry
from edsnlp.pipes.misc.tables import TablesMatcher

DEFAULT_CONFIG = dict(
    tables_pattern=None,
    sep_pattern=None,
    attr="TEXT",
    ignore_excluded=True,
)

create_component = registry.factory.register(
    "eds.tables",
    assigns=["doc.spans", "doc.ents"],
    deprecated=["tables"],
)(TablesMatcher)



import catalogue

try:
    from importlib.metadata import entry_points
except ImportError:
    from importlib_metadata import entry_points


def test_entrypoints():
    ep = entry_points()
    namespaces = ep.groups if hasattr(ep, "groups") else ep.keys()
    for ns in namespaces:
        if ns.startswith("spacy_") or ns.startswith("edsnlp_"):
            reg = catalogue.Registry(ns.split("_"), entry_points=True)
            reg.get_all()



import spacy
from pytest import mark

text = """COMPTE RENDU D'HOSPITALISATION du 11/07/2018 au 12/07/2018
MOTIF D'HOSPITALISATION
Monsieur Dupont Jean Michel, de sexe masculin, âgée de 39 ans, née le 23/11/1978,
a été hospitalisé du 11/08/2019 au 17/08/2019 pour une quinte de toux.

ANTÉCÉDENTS
Antécédents médicaux :
Premier épisode: il a été hospitalisé pour asthme en mai 2018."""


@mark.parametrize("use_sections", [True, False])
def test_reason(lang, use_sections):
    nlp = spacy.blank(lang)
    # Extraction d'entités nommées
    nlp.add_pipe(
        "eds.matcher",
        config=dict(
            terms=dict(
                respiratoire=[
                    "asthmatique",
                    "asthme",
                    "toux",
                ]
            )
        ),
    )
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.reason", config=dict(use_sections=use_sections))
    nlp.remove_pipe("eds.reason")
    nlp.add_pipe("eds.sections")
    nlp.add_pipe("eds.reason", config=dict(use_sections=use_sections))

    doc = nlp(text)
    reason = doc.spans["reasons"][0]
    entities = reason._.ents_reason

    assert entities[0].label_ == "respiratoire"
    assert reason._.is_reason
    assert doc.ents[1]._.is_reason is not use_sections



import pytest
from pytest import mark
from spacy.tokens import Span

import edsnlp


@mark.parametrize(
    "ner_mode,window",
    [
        ("independent", 1),
        ("joint", 0),
        ("joint", 5),
        ("marginal", 0),
    ],
)
def test_ner(ner_mode, window):
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        "eds.transformer",
        name="transformer",
        config=dict(
            model="prajjwal1/bert-tiny",
            window=128,
            stride=96,
        ),
    )
    nlp.add_pipe(
        "eds.ner_crf",
        name="ner",
        config=dict(
            embedding=nlp.get_pipe("transformer"),
            mode=ner_mode,
            target_span_getter=["ents", "ner-preds"],
            window=window,
        ),
    )

    ner = nlp.get_pipe("ner")
    ner.update_labels([])
    doc = nlp(
        "L'aîné eut le Moulin, le second eut l'âne, et le plus jeune n'eut que le Chat."
    )
    ner.labels = ["LOC", "ORG"]
    # doc[0:2], doc[4:5], doc[6:8], doc[9:11], doc[13:16], doc[20:21]
    doc.ents = [
        Span(doc, 0, 2, "PERSON"),
        Span(doc, 4, 5, "GIFT"),
        Span(doc, 6, 8, "PERSON"),
        Span(doc, 9, 11, "GIFT"),
        Span(doc, 13, 16, "PERSON"),
        Span(doc, 20, 21, "GIFT"),
    ]
    with pytest.warns() as record:
        nlp.post_init([doc])
    assert len(record) == 1
    assert record[0].message.args[0] == (
        "The labels inferred from the data are different from the labels passed to "
        "the component. Differing labels are ['GIFT', 'LOC', 'ORG', 'PERSON']"
    )

    ner = nlp.get_pipe("ner")
    ner.update_labels(["PERSON", "GIFT"])
    batch = ner.prepare_batch([doc], supervision=True)
    batch = ner.module_forward(batch)

    list(ner.pipe([doc]))

    assert batch["loss"] is not None



import re
import warnings
from collections import defaultdict
from functools import lru_cache
from operator import attrgetter
from typing import Any, Dict, List, Optional, Tuple, Union

import pydantic
from confit import VisibleDeprecationWarning
from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher, create_span
from edsnlp.matchers.utils import get_text
from edsnlp.pipes.base import BaseNERComponent, SpanSetterArg
from edsnlp.utils.collections import flatten_once

from . import models


@lru_cache(64)
def get_window(
    doclike: Union[Doc, Span], window: Tuple[int, int], limit_to_sentence: bool
):
    start_limit = doclike.sent.start if limit_to_sentence else 0
    end_limit = doclike.sent.end if limit_to_sentence else len(doclike.doc)

    start = max(doclike.start + window[0], start_limit)
    end = min(doclike.end + window[1], end_limit)

    return doclike.doc[start:end]


class ContextualMatcher(BaseNERComponent):
    """
    Allows additional matching in the surrounding context of the main match group,
    for qualification/filtering.

    Parameters
    ----------
    nlp : PipelineProtocol
        spaCy `Language` object.
    name : Optional[str]
        The name of the pipe
    patterns : Union[Dict[str, Any], List[Dict[str, Any]]]
        The configuration dictionary
    assign_as_span : bool
        Whether to store eventual extractions defined via the `assign` key as Spans
        or as string
    attr : str
        Attribute to match on, eg `TEXT`, `NORM`, etc.
    ignore_excluded : bool
        Whether to skip excluded tokens during matching.
    ignore_space_tokens : bool
        Whether to skip space tokens during matching.
    alignment_mode : str
        Overwrite alignment mode.
    regex_flags : Union[re.RegexFlag, int]
        RegExp flags to use when matching, filtering and assigning (See
        [here](https://docs.python.org/3/library/re.html#flags))
    include_assigned : bool
        Whether to include (eventual) assign matches to the final entity
    label_name : Optional[str]
        Deprecated, use `label` instead. The label to assign to the matched entities
    label : str
        The label to assign to the matched entities
    span_setter : SpanSetterArg
        How to set matches on the doc
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: Optional[str] = "contextual_matcher",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]],
        assign_as_span: bool = False,
        alignment_mode: str = "expand",
        attr: str = "NORM",
        regex_flags: Union[re.RegexFlag, int] = 0,
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        include_assigned: bool = False,
        label_name: Optional[str] = None,
        label: Optional[str] = None,
        span_setter: SpanSetterArg = {"ents": True},
    ):
        if label is None and label_name is not None:
            warnings.warn(
                "`label_name` is deprecated, use `label` instead.",
                VisibleDeprecationWarning,
            )
            label = label_name
        if label is None:
            raise ValueError("`label` parameter is required.")
        self.label = label

        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

        self.attr = attr
        self.assign_as_span = assign_as_span
        self.ignore_excluded = ignore_excluded
        self.ignore_space_tokens = ignore_space_tokens
        self.alignment_mode = alignment_mode
        self.regex_flags = regex_flags
        self.include_assigned = include_assigned

        # Configuration parsing
        patterns = pydantic.parse_obj_as(models.FullConfig, patterns)
        self.patterns = {pattern.source: pattern for pattern in patterns}

        # Matchers for the anchors
        self.phrase_matcher = EDSPhraseMatcher(
            nlp.vocab,
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
        )
        self.regex_matcher = RegexMatcher(
            attr=attr,
            flags=regex_flags,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
            alignment_mode=alignment_mode,
        )

        self.phrase_matcher.build_patterns(
            nlp=nlp,
            terms={
                source: {
                    "patterns": p.terms,
                }
                for source, p in self.patterns.items()
            },
        )
        self.regex_matcher.build_patterns(
            regex={
                source: {
                    "regex": p.regex,
                    "attr": p.regex_attr,
                    "flags": p.regex_flags,
                }
                for source, p in self.patterns.items()
            }
        )

        self.exclude_matchers = defaultdict(
            list
        )  # Will contain all the exclusion matchers
        self.assign_matchers = defaultdict(list)  # Will contain all the assign matchers

        # Will contain the reduce mode (for each source and assign matcher)
        self.reduce_mode = {}

        # Will contain the name of the assign matcher from which
        # entity will be replaced (for each source)
        self.replace_key = {}

        for source, p in self.patterns.items():
            p = p.dict()

            for exclude in p["exclude"]:
                exclude_matcher = RegexMatcher(
                    attr=exclude["regex_attr"] or p["regex_attr"] or self.attr,
                    flags=exclude["regex_flags"]
                    or p["regex_flags"]
                    or self.regex_flags,
                    ignore_excluded=ignore_excluded,
                    ignore_space_tokens=ignore_space_tokens,
                    alignment_mode="expand",
                )

                exclude_matcher.build_patterns(regex={"exclude": exclude["regex"]})

                self.exclude_matchers[source].append(
                    dict(
                        matcher=exclude_matcher,
                        window=exclude["window"],
                        limit_to_sentence=exclude["limit_to_sentence"],
                    )
                )

            replace_key = None

            for assign in p["assign"]:
                assign_matcher = RegexMatcher(
                    attr=assign["regex_attr"] or p["regex_attr"] or self.attr,
                    flags=assign["regex_flags"] or p["regex_flags"] or self.regex_flags,
                    ignore_excluded=ignore_excluded,
                    ignore_space_tokens=ignore_space_tokens,
                    alignment_mode=alignment_mode,
                    span_from_group=True,
                )

                assign_matcher.build_patterns(
                    regex={assign["name"]: assign["regex"]},
                )

                self.assign_matchers[source].append(
                    dict(
                        name=assign["name"],
                        matcher=assign_matcher,
                        window=assign["window"],
                        limit_to_sentence=assign["limit_to_sentence"],
                        replace_entity=assign["replace_entity"],
                        reduce_mode=assign["reduce_mode"],
                    )
                )

                if assign["replace_entity"]:
                    # We know that there is only one assign name
                    # with `replace_entity==True`
                    # from PyDantic validation
                    replace_key = assign["name"]

            self.replace_key[source] = replace_key

            self.reduce_mode[source] = {
                d["name"]: d["reduce_mode"] for d in self.assign_matchers[source]
            }

        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        if not Span.has_extension("assigned"):
            Span.set_extension("assigned", default=dict())
        if not Span.has_extension("source"):
            Span.set_extension("source", default=None)

    def filter_one(self, span: Span) -> Span:
        """
        Filter extracted entity based on the "exclusion filter" mentioned
        in the configuration

        Parameters
        ----------
        span : Span
            Span to filter

        Returns
        -------
        Optional[Span]
            None if the span was filtered, the span else
        """
        source = span.label_
        to_keep = True
        for matcher in self.exclude_matchers[source]:
            window = matcher["window"]
            limit_to_sentence = matcher["limit_to_sentence"]
            snippet = get_window(
                doclike=span,
                window=window,
                limit_to_sentence=limit_to_sentence,
            )

            if (
                next(
                    matcher["matcher"](snippet, as_spans=True),
                    None,
                )
                is not None
            ):
                to_keep = False
                logger.trace(f"Entity {span} was filtered out")
                break

        if to_keep:
            return span

    def assign_one(self, span: Span) -> Span:
        """
        Get additional information in the context
        of each entity. This function will populate two custom attributes:

        - `ent._.source`
        - `ent._.assigned`, a dictionary with all retrieved information

        Parameters
        ----------
        span : Span
            Span to enrich

        Returns
        -------
        Span
            Span with additional information
        """

        if span is None:
            yield from []
            return

        source = span.label_
        assigned_dict = models.AssignDict(reduce_mode=self.reduce_mode[source])
        replace_key = None

        for matcher in self.assign_matchers[source]:
            attr = self.patterns[source].regex_attr or matcher["matcher"].default_attr
            window = matcher["window"]
            limit_to_sentence = matcher["limit_to_sentence"]
            replace_entity = matcher["replace_entity"]  # Boolean

            snippet = get_window(
                doclike=span,
                window=window,
                limit_to_sentence=limit_to_sentence,
            )

            # Getting the matches
            assigned_list = list(matcher["matcher"].match(snippet))

            assigned_list = [
                (span, span, matcher["matcher"].regex[0][0])
                if not match.groups()
                else (
                    span,
                    create_span(
                        doclike=snippet,
                        start_char=match.start(0),
                        end_char=match.end(0),
                        key=matcher["matcher"].regex[0][0],
                        attr=matcher["matcher"].regex[0][2],
                        alignment_mode=matcher["matcher"].regex[0][5],
                        ignore_excluded=matcher["matcher"].regex[0][3],
                        ignore_space_tokens=matcher["matcher"].regex[0][4],
                    ),
                    matcher["matcher"].regex[0][0],
                )
                for (span, match) in assigned_list
            ]

            # assigned_list now contains tuples with
            # - the first element being the span extracted from the group
            # - the second element being the full match

            if not assigned_list:  # No match was found
                continue

            for assigned in assigned_list:
                if assigned is None:
                    continue
                if replace_entity:
                    replace_key = assigned[2]

                # Using he overrid `__setitem__` method from AssignDict here:
                assigned_dict[assigned[2]] = {
                    "span": assigned[1],  # Full span
                    "value_span": assigned[0],  # Span of the group
                    "value_text": get_text(
                        assigned[0],
                        attr=attr,
                        ignore_excluded=self.ignore_excluded,
                    ),  # Text of the group
                }
                logger.trace(f"Assign key {matcher['name']} matched on entity {span}")
        if replace_key is None and self.replace_key[source] is not None:
            # There should have been a replacement, but none was found
            # So we discard the entity
            return

        # Entity replacement
        if replace_key is not None:
            replacables = assigned_dict[replace_key]["span"]
            kept_ents = (
                replacables if isinstance(replacables, list) else [replacables]
            ).copy()

            if self.include_assigned:
                # We look for the closest
                closest = min(
                    kept_ents,
                    key=lambda e: abs(e.start - span.start),
                )
                kept_ents.remove(closest)

                expandables = list(
                    flatten_once(
                        [
                            a["span"]
                            for k, a in assigned_dict.items()
                            if k != replace_key
                        ]
                    )
                ) + [span, closest]

                closest = Span(
                    span.doc,
                    min(expandables, key=attrgetter("start")).start,
                    max(expandables, key=attrgetter("end")).end,
                    span.label_,
                )

                kept_ents.append(closest)
                kept_ents.sort(key=attrgetter("start"))

            for replaced in kept_ents:
                # Propagating attributes from the anchor
                replaced._.source = source
                replaced.label_ = self.label

        else:
            # Entity expansion
            expandables = [
                s
                for s in flatten_once([a["span"] for a in assigned_dict.values()])
                if s is not None
            ]

            if self.include_assigned and expandables:
                span = Span(
                    span.doc,
                    min(s.start for s in expandables + [span] if s is not None),
                    max(s.end for s in expandables + [span] if s is not None),
                    span.label_,
                )

            span._.source = source
            span.label_ = self.label
            kept_ents = [span]

        key = "value_span" if self.assign_as_span else "value_text"

        for idx, e in enumerate(kept_ents):
            e._.assigned = {
                k: v[key][idx]
                if ((k == replace_key) and self.reduce_mode[source][k] is None)
                else v[key]
                for k, v in assigned_dict.items()
            }

        yield from kept_ents

    def process_one(self, span):
        filtered = self.filter_one(span)
        yield from self.assign_one(filtered)

    def process(self, doc: Doc) -> List[Span]:
        """
        Process the document, looking for named entities.

        Parameters
        ----------
        doc : Doc
            spaCy Doc object

        Returns
        -------
        List[Span]
            List of detected spans.
        """

        matches = self.phrase_matcher(doc, as_spans=True)
        regex_matches = list(self.regex_matcher(doc, as_spans=True))

        spans = (*matches, *regex_matches)
        for span in spans:
            yield from self.process_one(span)

    def __call__(self, doc: Doc) -> Doc:
        """
        Adds spans to document.

        Parameters
        ----------
        doc:
            spaCy Doc object

        Returns
        -------
        doc:
            spaCy Doc object, annotated for extracted terms.
        """

        spans = list(self.process(doc))
        self.set_spans(doc, spans)
        return doc



from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol, registry


def remove_lowercase(doc: Doc):
    """
    Add case on the `NORM` custom attribute. Should always be applied first.

    Parameters
    ----------
    doc : Doc
        The spaCy `Doc` object.

    Returns
    -------
    Doc
        The document, with case put back in `NORM`.
    """

    for token in doc:
        token.norm_ = token.text

    return doc


@registry.factory.register(
    "eds.remove_lowercase",
    assigns=["token.norm"],
    deprecated=[
        "remove-lowercase",
        "eds.remove-lowercase",
    ],
)
def create_component(
    nlp: PipelineProtocol,
    name: str,
):
    """
    Add case on the `NORM` custom attribute. Should always be applied first.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : str
        The name of the component.
    """
    return remove_lowercase  # pragma: no cover



from edsnlp.data.standoff import (
    dump_standoff_file,
    parse_standoff_file,
    read_standoff,
    write_standoff,
)

load_from_brat = parse_standoff_file
export_to_brat = dump_standoff_file

read_brat = read_standoff
write_brat = write_standoff



from typing import List, Optional, Set, Union

from loguru import logger
from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanGetterArg, get_spans
from edsnlp.pipes.qualifiers.base import RuleBasedQualifier
from edsnlp.pipes.terminations import termination as default_termination
from edsnlp.utils.filter import consume_spans, filter_spans
from edsnlp.utils.inclusion import check_inclusion

from . import patterns


def family_getter(token: Union[Token, Span]) -> Optional[str]:
    if token._.family is True:
        return "FAMILY"
    elif token._.family is False:
        return "PATIENT"
    else:
        return None


class FamilyContextQualifier(RuleBasedQualifier):
    """
    The `eds.family` component uses a simple rule-based algorithm to detect spans that
    describe a family member (or family history) of the patient rather than the
    patient themself.

    Examples
    --------
    The following snippet matches a simple terminology, and checks the family context
    of the extracted entities. It is complete, and can be run _as is_.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    # Dummy matcher
    nlp.add_pipe(
        eds.matcher(terms=dict(douleur="douleur", osteoporose="ostéoporose")),
    )
    nlp.add_pipe(eds.family())

    text = (
        "Le patient est admis le 23 août 2021 pour une douleur au bras. "
        "Il a des antécédents familiaux d'ostéoporose"
    )

    doc = nlp(text)

    doc.ents
    # Out: (douleur, ostéoporose)

    doc.ents[0]._.family
    # Out: False

    doc.ents[1]._.family
    # Out: True
    ```

    Extensions
    ----------
    The `eds.family` component declares two extensions, on both `Span` and `Token`
    objects :

    1. The `family` attribute is a boolean, set to `True` if the component predicts
       that the span/token relates to a family member.
    2. The `family_` property is a human-readable string, computed from the `family`
       attribute. It implements a simple getter function that outputs `PATIENT` or
       `FAMILY`, depending on the value of `family`.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    name : Optional[str]
        The component name.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with the key 'term_attr'
        we can also add a key for each regex.
    family : Optional[List[str]]
        List of terms indicating family reference.
    termination : Optional[List[str]]
        List of syntagms termination terms.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Deprecated, use `span_getter` instead.

        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    explain : bool
        Whether to keep track of cues for each entity.
    use_sections : bool, by default `False`
        Whether to use annotated sections (namely `antécédents familiaux`).

    Authors and citation
    --------------------
    The `eds.family` component was developed by AP-HP's Data Science team.
    """

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "family",
        *,
        attr: str = "NORM",
        family: Optional[List[str]] = None,
        termination: Optional[List[str]] = None,
        use_sections: bool = True,
        span_getter: Optional[SpanGetterArg] = None,
        on_ents_only: Optional[Union[bool, str, List[str], Set[str]]] = None,
        explain: bool = False,
    ):
        terms = dict(
            family=patterns.family if family is None else family,
            termination=default_termination if termination is None else termination,
        )

        super().__init__(
            nlp=nlp,
            name=name,
            attr=attr,
            explain=explain,
            terms=terms,
            on_ents_only=on_ents_only,
            span_getter=span_getter,
        )

        self.sections = use_sections and (
            "eds.sections" in nlp.pipe_names or "sections" in nlp.pipe_names
        )
        if use_sections and not self.sections:
            logger.warning(
                "You have requested that the pipeline use annotations "
                "provided by the `section` pipeline, but it was not set. "
                "Skipping that step."
            )

    def set_extensions(self) -> None:
        super().set_extensions()
        for cls in (Token, Span):
            if not cls.has_extension("family"):
                cls.set_extension("family", default=None)

            if not cls.has_extension("family_"):
                cls.set_extension("family_", getter=family_getter)

        if not Span.has_extension("family_cues"):
            Span.set_extension("family_cues", default=[])

        if not Doc.has_extension("family"):
            Doc.set_extension("family", default=[])

    def process(self, doc: Doc) -> Doc:
        matches = self.get_matches(doc)

        terminations = [m for m in matches if m.label_ == "termination"]
        boundaries = self._boundaries(doc, terminations)

        # Removes duplicate matches and pseudo-expressions in one statement
        matches = filter_spans(matches, label_to_remove="pseudo")

        entities = list(get_spans(doc, self.span_getter))
        ents = None

        sections = []

        if self.sections:
            sections = [
                Span(doc, section.start, section.end, label="FAMILY")
                for section in doc.spans["sections"]
                if section.label_ == "antécédents familiaux"
            ]

        for start, end in boundaries:
            ents, entities = consume_spans(
                entities,
                filter=lambda s: check_inclusion(s, start, end),
                second_chance=ents,
            )

            sub_matches, matches = consume_spans(
                matches, lambda s: start <= s.start < end
            )

            sub_sections, sections = consume_spans(sections, lambda s: doc[start] in s)

            if self.on_ents_only and not ents:
                continue

            cues = [m for m in sub_matches if m.label_ == "family"]
            cues.extend(sub_sections)

            family = bool(cues)

            if not self.on_ents_only:
                for token in doc[start:end]:
                    token._.family = token._.family or family

            for ent in ents:
                ent._.family = family
                if self.explain:
                    ent._.family_cues += cues

        return doc



import datetime
from array import array
from collections import namedtuple
from decimal import Decimal

import numpy as np
import pytest
from pyspark import Row
from pyspark.sql.types import StringType

from edsnlp.utils.spark_dtypes import (
    PySparkPrettyPrinter,
    infer_schema,
    schema_warning,
    spark_interpret_dicts_as_rows,
)

mytype = namedtuple("mytype", ["a", "b", "c", "z"])


class CustomObj:
    def __init__(self, value):
        self.value = value


class HasUDT:
    __UDT__ = StringType()

    def __init__(self, value: str):
        self.value = value


def test_infer_schema():
    with spark_interpret_dicts_as_rows():
        result = infer_schema(
            {
                "null": None,
                "null_list": [None, None, None],
                "a": 1.0,
                "b": datetime.datetime(2021, 1, 1),
                "c": [
                    {
                        "d": "foo",
                        "e": 4,
                    },
                    {
                        "d": "bar",
                        "e": 6,
                    },
                ],
                "z": Decimal("1.0"),
                "struct": [mytype(1, 2, 3, 4), mytype(5, 6, 7, 8)],
                "custom": CustomObj(1),
                "udt": HasUDT("foo"),
                "rows": [Row(a=1, b=2), Row(a=3, b=4)],
                "tuples": [(1, 2), (3, 4)],
            }
        )
        assert PySparkPrettyPrinter().pformat(result) == (
            "T.StructType([\n"
            "    T.StructField('null', T.NullType(), True),\n"
            "    T.StructField('null_list', T.ArrayType(T.NullType(), True), True),\n"
            "    T.StructField('a', T.DoubleType(), True),\n"
            "    T.StructField('b', T.TimestampType(), True),\n"
            "    T.StructField('c', T.ArrayType(T.StructType([\n"
            "        T.StructField('d', T.StringType(), True),\n"
            "        T.StructField('e', T.LongType(), True)\n"
            "    ]), True), True),\n"
            "    T.StructField('z', T.DecimalType(), True),\n"
            "    T.StructField('struct', T.ArrayType(T.StructType([\n"
            "        T.StructField('a', T.LongType(), True),\n"
            "        T.StructField('b', T.LongType(), True),\n"
            "        T.StructField('c', T.LongType(), True),\n"
            "        T.StructField('z', T.LongType(), True)\n"
            "    ]), True), True),\n"
            "    T.StructField('custom', T.StructType([\n"
            "        T.StructField('value', T.LongType(), True)\n"
            "    ]), True),\n"
            "    T.StructField('udt', T.StringType(), True),\n"
            "    T.StructField('rows', T.ArrayType(T.StructType([\n"
            "        T.StructField('a', T.LongType(), True),\n"
            "        T.StructField('b', T.LongType(), True)\n"
            "    ]), True), True),\n"
            "    T.StructField('tuples', T.ArrayType(T.StructType([\n"
            "        T.StructField('_1', T.LongType(), True),\n"
            "        T.StructField('_2', T.LongType(), True)\n"
            "    ]), True), True)\n"
            "])"
        )

        with pytest.raises(TypeError):
            infer_schema({"numpy": np.array([1, 2, 3])})

        result = infer_schema(
            (1, 2, 3, 4, 5),
            names=["a", "b", "c"],
        )
        assert PySparkPrettyPrinter().pformat(result) == (
            "T.StructType([\n"
            "    T.StructField('a', T.LongType(), True),\n"
            "    T.StructField('b', T.LongType(), True),\n"
            "    T.StructField('c', T.LongType(), True),\n"
            "    T.StructField('_4', T.LongType(), True),\n"
            "    T.StructField('_5', T.LongType(), True)\n"
            "])"
        )

        result = infer_schema(
            {"field": array("d", [1.0, 2.0, 3.14])},
        )
        assert PySparkPrettyPrinter().pformat({"type": result}) == (
            "{'type': T.StructType([\n"
            "             T.StructField('field', T.ArrayType(T.DoubleType(), False), "
            "True)\n"
            "         ])}"
        )

        with pytest.raises(TypeError):
            infer_schema({"error_array": array("L", [1, 2, 3])})

        with pytest.warns(Warning) as warned:
            schema_warning(result)

        assert len(warned) == 1
        assert "The following schema was inferred" in warned[0].message.args[0]



from edsnlp.utils.examples import parse_example

d1v1 = "ABCD"
d1v2 = "BFEZ"
d2v1 = "0123"
d2v2 = "A1B3"
d2v3 = "ABC3"
d2v4 = "A990"
d2v5 = "A9AZ"
d2v6 = "0A12"

examples = [
    f"""1. Codification ADICAP : <ent text={d1v1 + d2v1}>{d1v1 + d2v1}</ent>.
    Une autre chose""",
    rf"""2. Codification ADICAP : <ent text={d1v1 + d2v2}>{d1v1 + d2v2}</ent>,\s
    <ent text={d1v1 + d2v3}>{d1v1 + d2v3}</ent>. Une autre chose""",
    f"""3. adicap : <ent text={d1v2 + d2v3}>{d1v2 + d2v3}</ent>,
    <ent text={d1v1 + d2v4}>{d1v1 + d2v4}</ent>. Une autre chose""",
    f"""4. Codification  : <ent text={d1v1 + d2v6}>{d1v1 + d2v6}</ent>.
    J'aime edsnlp. : {d1v2 + d2v3}.  Une autre chose""",
    f"""5. J'aime edsnlp. : {d1v2 + d2v5}.  Une autre chose""",
]


def test_scores(blank_nlp):
    if blank_nlp.lang == "eds":
        blank_nlp.add_pipe("eds.adicap")

        for example in examples:

            text, expected_entities = parse_example(example=example)

            doc = blank_nlp(text)

            for expected, ent in zip(expected_entities, doc.ents):
                assert text[expected.start_char : expected.end_char] == ent.text
                assert expected.modifiers[0].value == ent._.adicap.code
                assert len(ent._.adicap.dict()) > 0


def test_local_spelling(blank_nlp):
    blank_nlp.add_pipe("eds.adicap")
    txt = "Codification ADICAP : B.H.HP.A7A0 . Autre chose"
    assert blank_nlp(txt).ents[0]._.adicap.dict() == {
        "behaviour_type": "CANCER INVASIF",
        "code": "BHHPA7A0",
        "organ": "PROSTATE",
        "pathology": "PATHOLOGIE TUMORALE",
        "pathology_type": "ADENOCARCINOME INVASIF (SAI)",
        "sampling_mode": "BIOPSIE CHIRURGICALE",
        "technic": "HISTOLOGIE ET CYTOLOGIE PAR INCLUSION",
    }



from __future__ import annotations

from typing import Any, Callable, Iterable, Optional, Tuple, Union

import pandas as pd

from edsnlp import registry
from edsnlp.core.lazy_collection import LazyCollection
from edsnlp.data.base import BaseReader, BaseWriter
from edsnlp.data.converters import (
    FILENAME,
    get_dict2doc_converter,
    get_doc2dict_converter,
)
from edsnlp.utils.collections import dl_to_ld, flatten, ld_to_dl


class PandasReader(BaseReader):
    DATA_FIELDS = ("data",)

    def __init__(
        self,
        data: pd.DataFrame,
        **kwargs,
    ):
        assert isinstance(data, pd.DataFrame)
        self.data = data

        super().__init__(**kwargs)

    def read_main(self) -> Iterable[Tuple[Any, int]]:
        return ((item, 1) for item in dl_to_ld(dict(self.data)))

    def read_worker(self, fragments):
        return [task for task in fragments]


@registry.readers.register("pandas")
def from_pandas(
    data,
    converter: Optional[Union[str, Callable]] = None,
    **kwargs,
) -> LazyCollection:
    """
    The PandasReader (or `edsnlp.data.from_pandas`) handles reading from a table and
    yields documents. At the moment, only entities and attributes are loaded. Relations
    and events are not supported.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)
    doc_iterator = edsnlp.data.from_pandas(df, nlp=nlp, converter="omop")
    annotated_docs = nlp.pipe(doc_iterator)
    ```

    !!! note "Generator vs list"

        `edsnlp.data.from_pandas` returns a
        [LazyCollection][edsnlp.core.lazy_collection.LazyCollection].
        To iterate over the documents multiple times efficiently or to access them by
        index, you must convert it to a list

        ```{ .python .no-check }
        docs = list(edsnlp.data.from_pandas(df, converter="omop"))
        ```

    Parameters
    ----------
    data: pd.DataFrame
        Pandas object
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the rows of the DataFrame (represented as dicts)
        to Doc objects. These are documented on the [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.

    Returns
    -------
    LazyCollection
    """

    data = LazyCollection(reader=PandasReader(data))
    if converter:
        converter, kwargs = get_dict2doc_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)
    return data


class PandasWriter(BaseWriter):
    def __init__(self, dtypes: Optional[dict] = None):
        self.dtypes = dtypes

    def write_worker(self, records):
        # If write as jsonl, we will perform the actual writing in the `write` method
        for rec in records:
            if isinstance(rec, dict):
                rec.pop(FILENAME, None)
        return records, len(records)

    def write_main(self, fragments):
        import pandas as pd

        columns = ld_to_dl(flatten(fragments))
        res = pd.DataFrame(columns)
        return res.astype(self.dtypes) if self.dtypes else res


@registry.writers.register("pandas")
def to_pandas(
    data: Union[Any, LazyCollection],
    converter: Optional[Union[str, Callable]] = None,
    dtypes: Optional[dict] = None,
    **kwargs,
) -> pd.DataFrame:
    """
    `edsnlp.data.to_pandas` writes a list of documents as a pandas table.

    Example
    -------
    ```{ .python .no-check }

    import edsnlp

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(...)

    doc = nlp("My document with entities")

    edsnlp.data.to_pandas([doc], converter="omop")
    ```

    Parameters
    ----------
    data: Union[Any, LazyCollection],
        The data to write (either a list of documents or a LazyCollection).
    dtypes: Optional[dict]
        Dictionary of column names to dtypes. This is passed to `pd.DataFrame.astype`.
    converter: Optional[Union[str, Callable]]
        Converter to use to convert the documents to dictionary objects before storing
        them in the dataframe. These are documented on the
        [Converters](/data/converters) page.
    kwargs:
        Additional keyword arguments to pass to the converter. These are documented on
        the [Converters](/data/converters) page.
    """
    data = LazyCollection.ensure_lazy(data)
    if converter:
        converter, kwargs = get_doc2dict_converter(converter, kwargs)
        data = data.map(converter, kwargs=kwargs)

    return data.write(PandasWriter(dtypes))



TO_EXCLUDE = r"(?<!a )((\bacc\b)|anti.?coag|anti.?corps|buschke|(\bac\b)|(\bbio))"

main_pattern = dict(
    source="main",
    regex=[
        r"arthrites.{1,5}juveniles.{1,5}idiopa",
        r"myosite",
        r"myopathie.{1,5}inflammatoire",
        r"polyarthrite.{1,5}chronique.{1,5}evol",
        r"polymyosie",
        r"polyarthrites.{1,5}(rhizo|rhuma)",
        r"sclerodermie",
        r"connectivite",
        r"sarcoidose",
    ],
    exclude=dict(
        regex=[TO_EXCLUDE],
        window=(-7, 7),
    ),
    regex_attr="NORM",
)

lupus = dict(
    source="lupus",
    regex=[
        r"\blupus",
    ],
    regex_attr="NORM",
)

lupique = dict(
    source="lupique",
    regex=[
        r"\blupique",
    ],
    exclude=dict(
        regex=[TO_EXCLUDE],
        window=(-7, 7),
    ),
    regex_attr="NORM",
)

acronym = dict(
    source="acronyms",
    regex=[
        r"\bAJI\b",
        r"\bLED\b",
        r"\bPCE\b",
        r"\bCREST\b",
        r"\bPPR\b",
        r"\bMICI\b",
        r"\bMNAI\b",
    ],
    regex_attr="TEXT",
)

named_disease = dict(
    source="named_disease",
    regex=[
        r"libman.?lack",
        r"\bstill",
        r"felty",
        r"forestier.?certon",
        r"gou(g|j)erot",
        r"raynaud",
        r"thibierge.?weiss",
        r"sjogren",
        r"gou(g|j)erot.?sjogren",
    ],
    regex_attr="NORM",
)

default_patterns = [
    main_pattern,
    lupus,
    lupique,
    acronym,
    named_disease,
]



from typing import Iterable

from spacy.tokens import Doc, Span

from edsnlp.pipes.ner.scores.base_score import SimpleScoreMatcher


class SofaMatcher(SimpleScoreMatcher):
    def set_extensions(self):
        super().set_extensions()
        if not Span.has_extension("score_method"):
            Span.set_extension("score_method", default=None)

    def process(self, doc: Doc) -> Iterable[Span]:
        """
        Extracts, if available, the value of the score.
        Normalizes the score via the provided `self.score_normalization` method.

        Parameters
        ----------
        doc: Doc
            Document to process

        Returns
        -------
        ents: List[Span]
            List of spaCy's spans, with, if found, an added `score_value` extension
        """

        for ent in super().process(doc):
            assigned = ent._.assigned
            if not assigned:
                continue
            if assigned.get("method_max") is not None:
                method = "Maximum"
            elif assigned.get("method_24h") is not None:
                method = "24H"
            elif assigned.get("method_adm") is not None:
                method = "A l'admission"
            else:
                method = "Non précisée"

            normalized_value = self.score_normalization(assigned["value"])

            if normalized_value is not None:
                ent._.set(self.label, int(normalized_value))
                ent._.score_name = self.label
                ent._.score_method = method

                yield ent



from edsnlp.core import registry

from .negation import NegationQualifier

DEFAULT_CONFIG = dict(
    pseudo=None,
    preceding=None,
    preceding_regex=None,
    following=None,
    verbs=None,
    termination=None,
    attr="NORM",
    span_getter=None,
    on_ents_only=True,
    within_ents=False,
    explain=False,
)

create_component = registry.factory.register(
    "eds.negation",
    assigns=["span._.negation"],
    deprecated=["negation"],
)(NegationQualifier)



# cython: infer_types=True, profile=True
import re

from preshed.maps cimport map_clear, map_get, map_init, map_iter, map_set

from tqdm import tqdm

from preshed.maps cimport MapStruct, key_t
from spacy.matcher.phrasematcher cimport PhraseMatcher
from spacy.tokens.token cimport Token
from spacy.typedefs cimport attr_t
from spacy.vocab cimport Vocab

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.utils import Patterns, normalize_token_attr

def get_normalized_variant(doclike) -> str:
    tokens = [t.text + t.whitespace_ for t in doclike if not t._.excluded]
    variant = "".join(tokens)
    variant = variant.rstrip(" ")
    variant = re.sub(r"\s+", " ", variant)
    return variant


cdef class EDSPhraseMatcher(PhraseMatcher):
    """
    PhraseMatcher that allows to skip excluded tokens.
    Adapted from https://github.com/explosion/spaCy/blob/master/spacy/matcher/phrasematcher.pyx

    Parameters
    ----------
    vocab : Vocab
        spaCy vocabulary to match on.
    attr : str
        Default attribute to match on, by default "TEXT".
        Can be overridden in the `add` method.
        To match on a custom attribute, prepend the attribute name with `_`.
    ignore_excluded : bool, optional
        Whether to ignore excluded tokens, by default True
    ignore_space_tokens : bool, optional
        Whether to exclude tokens that have a "SPACE" tag, by default False
    """

    def __init__(self, Vocab vocab, attr="ORTH", ignore_excluded=True, ignore_space_tokens=False, validate=False):
        """Initialize the PhraseMatcher.

        vocab (Vocab): The shared vocabulary.
        attr (int / str): Token attribute to match on.
        validate (bool): Perform additional validation when patterns are added.

        DOCS: https://spacy.io/api/phrasematcher#init
        """
        super().__init__(vocab, attr, validate)

        if ignore_excluded:
            self.excluded_hash = vocab.strings['EXCLUDED']
        else:
            self.excluded_hash = 0

        if ignore_space_tokens:
            self.space_hash = vocab.strings['SPACE']
        else:
            self.space_hash = 0

        self.set_extensions()

    @staticmethod
    def set_extensions():
        if not Span.has_extension("normalized_variant"):
            Span.set_extension("normalized_variant", getter=get_normalized_variant)

    def build_patterns(self, nlp: PipelineProtocol, terms: Patterns, progress: bool = False):
        """
        Build patterns and adds them for matching.
        Helper function for pipelines using this matcher.

        Parameters
        ----------
        nlp : PipelineProtocol
            The instance of the spaCy language class.
        terms : Patterns
            Dictionary of label/terms, or label/dictionary of terms/attribute.
        progress: bool
            Whether to track progress when preprocessing terms
        """

        if not terms:
            terms = dict()

        matched_attr = normalize_token_attr(self.vocab.strings[self.attr])
        assert matched_attr is not None, "Unsupported attribute for matching"
        token_pipelines = [
            name
            for name, pipe in nlp.pipeline
            if name not in nlp.disabled and matched_attr in {
                normalize_token_attr(ass)
                for ass in nlp.get_pipe_meta(name).assigns
            }
        ]
        with nlp.select_pipes(enable=token_pipelines):
            for key, expressions in (tqdm(
                terms.items(),
                desc="Adding terms into the pipeline"
            ) if progress else terms.items()):
                if isinstance(expressions, dict):
                    attr = expressions.get("attr")
                    expressions = expressions.get("patterns")
                else:
                    attr = None
                if isinstance(expressions, str):
                    expressions = [expressions]
                patterns = list(nlp.pipe(expressions))
                self.add(key, patterns, attr)

    cdef void find_matches(self, Doc doc, int start_idx, int end_idx, vector[SpanC] *matches) nogil:
        cdef:
            MapStruct * current_node = self.c_map
            int start = 0
            int idx = start_idx
            int idy = start_idx
            key_t key
            void * value
            int i = 0
            SpanC ms
            void * result
        while idx < end_idx:
            while (
                (0 < self.space_hash == doc.c[idx].tag)
                or (0 < self.excluded_hash == doc.c[idx].tag)
            ):
                idx += 1
                if idx >= end_idx:
                    return
            start = idx
            # look for sequences from this position
            token = Token.get_struct_attr(&doc.c[idx], self.attr)
            result = map_get(current_node, token)
            if result:
                current_node = <MapStruct *> result
                idy = idx + 1
                while idy < end_idx:
                    result = map_get(current_node, self._terminal_hash)
                    if result:
                        i = 0
                        while map_iter(<MapStruct *> result, &i, &key, &value):
                            ms = make_spanstruct(key, start, idy)
                            matches.push_back(ms)

                    while idy < end_idx and (
                        (0 < self.space_hash == doc.c[idy].tag)
                        or (0 < self.excluded_hash == doc.c[idy].tag)
                    ):
                        idy += 1
                    inner_token = Token.get_struct_attr(&doc.c[idy], self.attr)
                    result = map_get(current_node, inner_token)
                    if result:
                        current_node = <MapStruct *> result
                        idy += 1
                    else:
                        break
                else:
                    # end of doc reached
                    result = map_get(current_node, self._terminal_hash)
                    if result:
                        i = 0
                        while map_iter(<MapStruct *> result, &i, &key, &value):
                            ms = make_spanstruct(key, start, idy)
                            matches.push_back(ms)
            current_node = self.c_map
            idx += 1

cdef SpanC make_spanstruct(attr_t label, int start, int end) nogil:
    cdef SpanC spanc
    spanc.label = label
    spanc.start = start
    spanc.end = end
    return spanc



from typing import List, Optional, Union

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.core.matcher.matcher import GenericMatcher
from edsnlp.pipes.misc.dates.factory import DEFAULT_CONFIG, DatesMatcher

from ...base import SpanSetterArg
from . import patterns as consult_regex


class ConsultationDatesMatcher(GenericMatcher):
    '''
    The `eds.consultation-dates` matcher consists of two main parts:

    - A **matcher** which finds mentions of _consultation events_ (more details below)
    - A **date parser** (see the corresponding pipe) that links a date to those events

    Examples
    --------
    !!! note

        The matcher has been built to run on **consultation notes**
        (`CR-CONS` at APHP), so please filter accordingly before proceeding.

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            lowercase=True,
            accents=True,
            quotes=True,
            pollution=False,
        ),
    )
    nlp.add_pipe(eds.consultation_dates())

    text = """
    XXX
    Objet : Compte-Rendu de Consultation du 03/10/2018.
    XXX
    """

    doc = nlp(text)

    doc.spans["consultation_dates"]
    # Out: [Consultation du 03/10/2018]

    doc.spans["consultation_dates"][0]._.consultation_date.to_datetime()
    # Out: DateTime(2018, 10, 3, 0, 0, 0, tzinfo=Timezone('Europe/Paris'))
    ```

    Extensions
    ----------
    The `eds.consultation_dates` pipeline declares one extension on the `Span` object:
    the `consultation_date` attribute, which is a Python `datetime` object.

    Parameters
    ----------
    nlp : PipelineProtocol
        Language pipeline object
    consultation_mention : Union[List[str], bool]
        List of RegEx for consultation mentions.

        - If `type==list`: Overrides the default list
        - If `type==bool`: Uses the default list of True, disable if False

        This list contains terms directly referring to consultations, such as
        "_Consultation du..._" or "_Compte rendu du..._". This list is the only one
        enabled by default since it is fairly precise and not error-prone.
    town_mention : Union[List[str], bool]
        List of RegEx for all AP-HP hospitals' towns mentions.

        - If `type==list`: Overrides the default list
        - If `type==bool`: Uses the default list of True, disable if False

        This list contains the towns of each AP-HP's hospital. Its goal is to fetch
        dates mentioned as "_Paris, le 13 décembre 2015_". It has a high recall but
        poor precision, since those dates can often be dates of letter redaction
        instead of consultation dates.
    document_date_mention : Union[List[str], bool]
        List of RegEx for document date.

        - If `type==list`: Overrides the default list
        - If `type==bool`: Uses the default list of True, disable if False

        This list contains expressions mentioning the date of creation/edition of a
        document, such as "_Date du rapport: 13/12/2015_" or "_Signé le 13/12/2015_".
        Like `town_mention` patterns, it has a high recall but is prone to errors since
        document date and consultation date aren't necessary similar.

    Authors and citation
    --------------------
    The `eds.consultation_dates` pipeline was developed by AP-HP's Data Science team.
    '''

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "consultation_dates",
        *,
        consultation_mention: Union[bool, List[str]] = True,
        town_mention: Union[bool, List[str]] = False,
        document_date_mention: Union[bool, List[str]] = False,
        attr: str = "NORM",
        ignore_excluded: bool = False,
        ignore_space_tokens: bool = False,
        label: str = "consultation_date",
        span_setter: SpanSetterArg = {"ents": True, "consultation_dates": True},
    ):
        logger.warning("This pipeline is still in beta")
        logger.warning(
            "This pipeline should ONLY be used on notes "
            "where `note_class_source_value == 'CR-CONS'`"
        )
        logger.warning(
            """This pipeline requires to use the normalizer pipeline with:
        lowercase=True,
        accents=True,
        quotes=True"""
        )

        if not (nlp.has_pipe("dates") and nlp.get_pipe("dates").on_ents_only is False):
            self.date_matcher = DatesMatcher(nlp, **DEFAULT_CONFIG)

        else:
            self.date_matcher = None

        if not consultation_mention:
            consultation_mention = []
        elif consultation_mention is True:
            consultation_mention = consult_regex.consultation_mention

        if not document_date_mention:
            document_date_mention = []
        elif document_date_mention is True:
            document_date_mention = consult_regex.document_date_mention

        if not town_mention:
            town_mention = []
        elif town_mention is True:
            town_mention = consult_regex.town_mention

        regex = dict(
            consultation_mention=consultation_mention,
            town_mention=town_mention,
            document_date_mention=document_date_mention,
        )
        self.label = label

        super().__init__(
            nlp=nlp,
            name=name,
            regex=regex,
            terms=dict(),
            attr=attr,
            ignore_excluded=ignore_excluded,
            ignore_space_tokens=ignore_space_tokens,
            term_matcher="exact",
            term_matcher_config=dict(),
            span_setter=span_setter,
        )

        self.set_extensions()

    def set_extensions(self) -> None:
        super().set_extensions()
        if not Span.has_extension(self.label):
            Span.set_extension(self.label, default=None)

    def process(self, doc: Doc) -> List[Span]:
        """
        Finds entities

        Parameters
        ----------
        doc: spaCy Doc object

        Returns
        -------
        doc: Doc
            spaCy Doc object with additional
            `doc.spans['consultation_dates]` `SpanGroup`
        """

        matches = list(super().process(doc))

        self.date_matcher.span_getter = lambda d: [m.sent for m in matches]
        dates = [s for s in self.date_matcher.process(doc) if s.label_ == "date"]
        self.date_matcher.span_getter = None

        for match in matches:
            # Looking for a date
            # - In the same sentence
            # - Not less than 10 tokens AFTER the consultation mention
            matching_dates = [
                date
                for date in dates
                if (
                    (match.sent == date.sent)
                    and (date.start > match.start)
                    and (date.start - match.end <= 10)
                )
            ]

            if matching_dates:
                # We keep the first mention of a date
                kept_date = min(matching_dates, key=lambda d: d.start)
                span = doc[match.start : kept_date.end]
                span.label_ = self.label
                span._.consultation_date = kept_date._.date

                yield span



import os
from enum import Enum
from functools import wraps
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import safetensors.torch
import torch
from spacy.tokens import Doc

from edsnlp.pipes.base import BaseComponent, BaseComponentMeta
from edsnlp.utils.collections import batch_compress_dict, batchify, decompress_dict

BatchInput = TypeVar("BatchInput", bound=Dict[str, Any])
BatchOutput = TypeVar("BatchOutput", bound=Dict[str, Any])
Scorer = Callable[[Sequence[Tuple[Doc, Doc]]], Union[float, Dict[str, Any]]]

ALL_CACHES = object()
_caches = {}


class CacheEnum(str, Enum):
    preprocess = "preprocess"
    collate = "collate"
    forward = "forward"


def hash_batch(batch):
    if isinstance(batch, list):
        return hash(tuple(id(item) for item in batch))
    elif not isinstance(batch, dict):
        return id(batch)
    if "__batch_hash__" in batch:
        return batch["__batch_hash__"]
    batch_hash = hash((tuple(batch.keys()), tuple(map(hash_batch, batch.values()))))
    batch["__batch_hash__"] = batch_hash
    return batch_hash


def cached_preprocess(fn):
    @wraps(fn)
    def wrapped(self: "TorchComponent", doc: Doc):
        if self._current_cache_id is None:
            return fn(self, doc)
        cache_key = (
            "preprocess",
            f"{type(self).__name__}<{id(self)}>",
            id(doc),
        )
        cache = _caches[self._current_cache_id]
        if cache_key in cache:
            return cache[cache_key]
        res = fn(self, doc)
        cache[cache_key] = res
        return res

    return wrapped


def cached_preprocess_supervised(fn):
    @wraps(fn)
    def wrapped(self: "TorchComponent", doc: Doc):
        if self._current_cache_id is None:
            return fn(self, doc)
        cache_key = (
            "preprocess_supervised",
            f"{type(self).__name__}<{id(self)}>",
            id(doc),
        )
        cache = _caches[self._current_cache_id]
        if cache_key in cache:
            return cache[cache_key]
        res = fn(self, doc)
        cache[cache_key] = res
        return res

    return wrapped


def cached_collate(fn):
    @wraps(fn)
    def wrapped(self: "TorchComponent", batch: Dict):
        if self._current_cache_id is None:
            return fn(self, batch)
        batch_hash = hash_batch(batch)
        cache_key = (
            "collate",
            f"{type(self).__name__}<{id(self)}>",
            batch_hash,
        )
        cache = _caches[self._current_cache_id]
        if cache_key in cache:
            return cache[cache_key]
        res = fn(self, batch)
        res["__batch_hash__"] = batch_hash
        cache[cache_key] = res
        return res

    return wrapped


def cached_forward(fn):
    @wraps(fn)
    def wrapped(self: "TorchComponent", batch):
        # Convert args and kwargs to a dictionary matching fn signature
        if self._current_cache_id is None:
            return fn(self, batch)
        cache_key = (
            "forward",
            f"{type(self).__name__}<{id(self)}>",
            hash_batch(batch),
        )
        cache = _caches[self._current_cache_id]
        if cache_key in cache:
            return cache[cache_key]
        res = fn(self, batch)
        cache[cache_key] = res
        return res

    return wrapped


def cached_batch_to_device(fn):
    @wraps(fn)
    def wrapped(self: "TorchComponent", batch, device):
        # Convert args and kwargs to a dictionary matching fn signature
        if self._current_cache_id is None:
            return fn(self, batch, device)
        cache_key = (
            "batch_to_device",
            f"{type(self).__name__}<{id(self)}>",
            hash_batch(batch),
        )
        cache = _caches[self._current_cache_id]
        if cache_key in cache:
            return cache[cache_key]
        res = fn(self, batch, device)
        cache[cache_key] = res
        return res

    return wrapped


class TorchComponentMeta(BaseComponentMeta):
    def __new__(mcs, name, bases, class_dict):
        if "preprocess" in class_dict:
            class_dict["preprocess"] = cached_preprocess(class_dict["preprocess"])
        if "preprocess_supervised" in class_dict:
            class_dict["preprocess_supervised"] = cached_preprocess_supervised(
                class_dict["preprocess_supervised"]
            )
        if "collate" in class_dict:
            class_dict["collate"] = cached_collate(class_dict["collate"])
        if "batch_to_device" in class_dict:
            class_dict["batch_to_device"] = cached_batch_to_device(
                class_dict["batch_to_device"]
            )
        if "forward" in class_dict:
            class_dict["forward"] = cached_forward(class_dict["forward"])

        return super().__new__(mcs, name, bases, class_dict)


class TorchComponent(
    BaseComponent,
    torch.nn.Module,
    Generic[BatchOutput, BatchInput],
    metaclass=TorchComponentMeta,
):
    """
    A TorchComponent is a Component that can be trained and inherits `torch.nn.Module`.
    You can use it either as a torch module inside a more complex neural network, or as
    a standalone component in a [Pipeline][edsnlp.core.pipeline.Pipeline].

    In addition to the methods of a torch module, a TorchComponent adds a few methods to
    handle preprocessing and collating features, as well as caching intermediate results
    for components that share a common subcomponent.
    """

    call_super_init = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._current_cache_id = None

    def enable_cache(self, cache_id="default"):
        self._current_cache_id = cache_id
        _caches.setdefault(cache_id, {})
        for name, component in self.named_component_children():
            if hasattr(component, "enable_cache"):
                component.enable_cache(cache_id)

    def disable_cache(self, cache_id=ALL_CACHES):
        if cache_id is ALL_CACHES:
            _caches.clear()
        else:
            if cache_id in _caches:
                del _caches[cache_id]
        self._current_cache_id = None
        for name, component in self.named_component_children():
            if hasattr(component, "disable_cache"):
                component.disable_cache(cache_id)

    @property
    def device(self):
        return next((p.device for p in self.parameters()), torch.device("cpu"))

    def named_component_children(self):
        for name, module in self.named_children():
            if isinstance(module, TorchComponent):
                yield name, module

    def named_component_modules(self):
        for name, module in self.named_modules():
            if isinstance(module, TorchComponent):
                yield name, module

    def post_init(self, gold_data: Iterable[Doc], exclude: Set[str]):
        """
        This method completes the attributes of the component, by looking at some
        documents. It is especially useful to build vocabularies or detect the labels
        of a classification task.

        Parameters
        ----------
        gold_data: Iterable[Doc]
            The documents to use for initialization.
        exclude: Optional[set]
            The names of components to exclude from initialization.
            This argument will be gradually updated  with the names of initialized
            components
        """
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        exclude.add(repr_id)
        for name, component in self.named_component_children():
            if hasattr(component, "post_init"):
                component.post_init(gold_data, exclude=exclude)

    def preprocess(self, doc: Doc) -> Dict[str, Any]:
        """
        Preprocess the document to extract features that will be used by the
        neural network to perform its predictions.

        Parameters
        ----------
        doc: Doc
            Document to preprocess

        Returns
        -------
        Dict[str, Any]
            Dictionary (optionally nested) containing the features extracted from
            the document.
        """
        return {
            name: component.preprocess(doc)
            for name, component in self.named_component_children()
        }

    def collate(self, batch: Dict[str, Any]) -> BatchInput:
        """
        Collate the batch of features into a single batch of tensors that can be
        used by the forward method of the component.

        Parameters
        ----------
        batch: Dict[str, Any]
            Batch of features

        Returns
        -------
        BatchInput
            Dictionary (optionally nested) containing the collated tensors
        """
        return {
            name: component.collate(batch[name])
            for name, component in self.named_component_children()
        }

    def batch_to_device(
        self,
        batch: BatchInput,
        device: Optional[Union[str, torch.device]],
    ) -> BatchInput:
        """
        Move the batch of tensors to the specified device.

        Parameters
        ----------
        batch: BatchInput
            Batch of tensors
        device: Optional[Union[str, torch.device]]
            Device to move the tensors to

        Returns
        -------
        BatchInput
        """
        return {
            name: (
                (value.to(device) if device is not None else value)
                if hasattr(value, "to")
                else getattr(self, name).batch_to_device(value, device=device)
                if hasattr(self, name)
                and hasattr(getattr(self, name), "batch_to_device")
                else value
            )
            for name, value in batch.items()
        }

    def forward(self, batch: BatchInput) -> BatchOutput:
        """
        Perform the forward pass of the neural network.

        Parameters
        ----------
        batch: BatchInput
            Batch of tensors (nested dictionary) computed by the collate method

        Returns
        -------
        BatchOutput
        """
        raise NotImplementedError()

    def module_forward(self, *args, **kwargs):
        """
        This is a wrapper around `torch.nn.Module.__call__` to avoid conflict
        with the components `__call__` method.
        """
        return torch.nn.Module.__call__(self, *args, **kwargs)

    def prepare_batch(
        self,
        docs: Sequence[Doc],
        supervision: bool = False,
        device: Optional[Union[str, torch.device]] = None,
    ) -> BatchInput:
        """
        Convenience method to preprocess a batch of documents and collate them
        Features corresponding to the same path are grouped together in a list,
        under the same key.

        Parameters
        ----------
        docs: Sequence[Doc]
            Batch of documents
        supervision: bool
            Whether to extract supervision features or not
        device: Optional[Union[str, torch.device]]
            Device to move the tensors to

        Returns
        -------
        Dict[str, Sequence[Any]]
        """
        batch = [
            (self.preprocess_supervised(doc) if supervision else self.preprocess(doc))
            for doc in docs
        ]
        batch = decompress_dict(list(batch_compress_dict(batch)))
        batch = self.collate(batch)
        batch = self.batch_to_device(batch, device=device)
        return batch

    def batch_process(self, docs: Sequence[Doc]) -> Sequence[Doc]:
        """
        Process a batch of documents using the neural network.
        This differs from the `pipe` method in that it does not return an
        iterator, but executes the component on the whole batch at once.

        Parameters
        ----------
        docs: Sequence[Doc]
            Batch of documents

        Returns
        -------
        Sequence[Doc]
            Batch of updated documents
        """
        with torch.no_grad():
            batch = self.prepare_batch(docs, device=self.device)
            if hasattr(self, "compiled"):
                res = self.compiled(batch)
            else:
                res = self.module_forward(batch)
            docs = self.postprocess(docs, res)
            return docs

    def postprocess(self, docs: Sequence[Doc], batch: BatchOutput) -> Sequence[Doc]:
        """
        Update the documents with the predictions of the neural network.
        By default, this is a no-op.

        Parameters
        ----------
        docs: Sequence[Doc]
            Batch of documents
        batch: BatchOutput
            Batch of predictions, as returned by the forward method

        Returns
        -------
        Sequence[Doc]
        """
        return docs

    # Same as preprocess but with gold supervision data
    def preprocess_supervised(self, doc: Doc) -> Dict[str, Any]:
        """
        Preprocess the document to extract features that will be used by the
        neural network to perform its training.
        By default, this returns the same features as the `preprocess` method.

        Parameters
        ----------
        doc: Doc
            Document to preprocess

        Returns
        -------
        Dict[str, Any]
            Dictionary (optionally nested) containing the features extracted from
            the document.
        """
        return self.preprocess(doc)

    def pipe(self, docs: Iterable[Doc], batch_size=1) -> Iterable[Doc]:
        """
        Applies the component on a collection of documents. It is recommended to use
        the [`Pipeline.pipe`][edsnlp.core.pipeline.Pipeline.pipe] method instead of this
        one to apply a pipeline on a collection of documents, to benefit from the
        caching of intermediate results.

        Parameters
        ----------
        docs: Iterable[Doc]
            Input docs
        batch_size: int
            Batch size to use when making batched to be process at once
        """
        for batch in batchify(docs, batch_size=batch_size):
            yield from self.batch_process(batch)

    def __call__(self, doc: Doc) -> Doc:
        """
        Applies the component on a single doc.
        For multiple documents, prefer batch processing via the
        [batch_process][edsnlp.core.torch_component.TorchComponent.batch_process]
        method. In general, prefer the [Pipeline][edsnlp.core.pipeline.Pipeline] methods

        Parameters
        ----------
        doc: Doc

        Returns
        -------
        Doc
        """
        return self.batch_process([doc])[0]

    def to_disk(self, path, *, exclude: Optional[Set[str]]):
        if object.__repr__(self) in exclude:
            return
        exclude.add(object.__repr__(self))
        overrides = {}
        for name, component in self.named_component_children():
            if hasattr(component, "to_disk"):
                pipe_overrides = component.to_disk(path / name, exclude=exclude)
                if pipe_overrides:
                    overrides[name] = pipe_overrides
        tensor_dict = {
            n: p
            for n, p in (*self.named_parameters(), *self.named_buffers())
            if object.__repr__(p) not in exclude
        }
        os.makedirs(path, exist_ok=True)
        safetensors.torch.save_file(tensor_dict, path / "parameters.safetensors")
        exclude.update(object.__repr__(p) for p in tensor_dict.values())
        return overrides

    def from_disk(self, path, exclude: Optional[Set[str]]):
        if object.__repr__(self) in exclude:
            return
        exclude.add(object.__repr__(self))
        for name, component in self.named_component_children():
            if hasattr(component, "from_disk"):
                component.from_disk(path / name, exclude=exclude)
        tensor_dict = safetensors.torch.load_file(path / "parameters.safetensors")
        self.load_state_dict(tensor_dict, strict=False)



import pytest
from confit import validate_arguments
from confit.errors import ConfitValidationError

from edsnlp.utils.bindings import BINDING_GETTERS, BINDING_SETTERS, QualifiersArg


def test_qualifier_validation():
    @validate_arguments
    def fn(arg: QualifiersArg):
        return arg

    assert fn("_.negated") == {"_.negated": True}
    assert fn(["_.negated", "_.event"]) == {"_.negated": True, "_.event": True}
    assert fn({"_.negated": True, "_.event": "DATE"}) == {
        "_.negated": True,
        "_.event": ["DATE"],
    }

    callback = lambda x: x  # noqa: E731

    assert fn(callback) is callback

    with pytest.raises(ConfitValidationError):
        fn(1)

    with pytest.raises(ConfitValidationError):
        fn({"_.negated": 1})


def test_bindings():
    class custom:
        def __init__(self, value):
            self.value = value

    obj = custom([custom(1), custom(2)])
    assert BINDING_GETTERS["value[0].value"](obj) == 1
    assert BINDING_GETTERS[("value[0].value", 1)](obj) is True
    BINDING_SETTERS[("value[1].value", 3)](obj)
    assert obj.value[1].value == 3



from __future__ import annotations

import warnings
from collections import defaultdict
from enum import Enum
from typing import Callable, Dict, Iterable, List, Optional, Sequence, Set, Union

import torch
from spacy.tokens import Doc, Span
from typing_extensions import Literal, NotRequired, TypedDict

from edsnlp.core.pipeline import Pipeline
from edsnlp.core.torch_component import TorchComponent
from edsnlp.pipes.base import BaseNERComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    BatchInput,
    WordEmbeddingComponent,
)
from edsnlp.pipes.trainable.layers.crf import MultiLabelBIOULDecoder
from edsnlp.utils.filter import align_spans, filter_spans
from edsnlp.utils.span_getters import (
    SpanGetterArg,
    SpanGetterMapping,
    SpanSetter,
    SpanSetterArg,
    get_spans,
)
from edsnlp.utils.torch import make_windows

NERBatchInput = TypedDict(
    "NERBatchInput",
    {
        "embedding": BatchInput,
        "targets": NotRequired[torch.Tensor],
        "window_indices": NotRequired[torch.Tensor],
        "window_indexer": NotRequired[torch.Tensor],
    },
)
NERBatchOutput = TypedDict(
    "NERBatchOutput",
    {
        "loss": Optional[torch.Tensor],
        "tags": Optional[torch.Tensor],
    },
)


class CRFMode(str, Enum):
    independent = "independent"
    joint = "joint"
    marginal = "marginal"


class TrainableNerCrf(TorchComponent[NERBatchOutput, NERBatchInput], BaseNERComponent):
    """
    The `eds.ner_crf` component is a general purpose trainable named entity recognizer.
    It can extract:

    - flat entities
    - overlapping entities of different labels

    However, at the moment, the model cannot currently extract entities that are
    nested inside larger entities of the same label.

    It is based on a CRF (Conditional Random Field) layer and should therefore work
    well on dataset composed of entities will ill-defined boundaries. We offer a
    compromise between speed and performance by allowing the user to specify a window
    size for the CRF layer. The smaller the window, the faster the model will be, but
    at the cost of degraded performance.

    The pipeline assigns both `doc.ents` (in which overlapping entities are filtered
    out) and `doc.spans`. These destinations can be inferred from the
    `target_span_getter` parameter, combined with the `post_init` step.

    Architecture
    ------------
    The model performs token classification using the BIOUL (Begin, Inside, Outside,
    Unary, Last) tagging scheme. To extract overlapping entities, each label has its
    own tag sequence, so the model predicts `n_labels` sequences of O, I, B, L, U tags.
    The architecture is displayed in the figure below.

    To enforce the tagging scheme, (ex: I cannot follow O but only B, ...), we use a
    stack of CRF (Conditional Random Fields) layers, one per label during both training
    and prediction.

    <figure markdown>
      ![Nested NER architecture](/assets/images/edsnlp-ner.svg)
      <figcaption>Nested NER architecture</figcaption>
    </figure>

    Examples
    --------
    Let us define a pipeline composed of a transformer, and a NER component.

    ```{ .python }
    from pathlib import Path

    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.ner_crf(
            embedding=eds.transformer(
                model="prajjwal1/bert-tiny",
                window=128,
                stride=96,
            ),
            mode="joint",
            target_span_getter=["ents", "ner-preds"],
            window=10,
        ),
        name="ner"
    )
    ```

    To train the model, refer to the [Training](/tutorials/make-a-training-script)
    tutorial.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        Name of the component
    embedding : WordEmbeddingComponent
        The word embedding component
    target_span_getter : SpanGetterArg
        Method to call to get the gold spans from a document, for scoring or training.
        By default, takes all entities in `doc.ents`, but we recommend you specify
        a given span group name instead.
    labels : List[str]
        The labels to predict. The labels can also be inferred from the data
        during `nlp.post_init(...)`
    span_setter : Optional[SpanSetterArg]
        The span setter to use to set the predicted spans on the Doc object. If None,
        the component will infer the span setter from the target_span_getter config.
    infer_span_setter : Optional[bool]
        Whether to complete the span setter from the target_span_getter config.
        False by default, unless the span_setter is None.
    mode : Literal["independent", "joint", "marginal"]
        The CRF mode to use : independent, joint or marginal
    window : int
        The window size to use for the CRF. If 0, will use the whole document, at
        the cost of a longer computation time. If 1, this is equivalent to assuming
        that the tags are independent and will the component be faster, but with
        degraded performance. Empirically, we found that a window size of 10 or 20
        works well.
    stride : Optional[int]
        The stride to use for the CRF windows. Defaults to `window - 2`.

    Authors and citation
    --------------------
    The `eds.ner_crf` pipeline was developed by AP-HP's Data Science team.

    The deep learning model was adapted from [@wajsburt:tel-03624928].
    """

    span_setter: SpanSetter

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: Optional[str] = "ner_crf",
        *,
        embedding: WordEmbeddingComponent,
        target_span_getter: SpanGetterArg = {"ents": True},
        labels: Optional[List[str]] = None,
        span_setter: Optional[SpanSetterArg] = None,
        infer_span_setter: Optional[bool] = None,
        mode: Literal["independent", "joint", "marginal"],
        window: int = 40,
        stride: Optional[int] = None,
    ):
        if (
            isinstance(target_span_getter, dict) and "labels" in target_span_getter
        ) and labels is not None:
            raise ValueError(
                "You cannot set both the `labels` key of the `target_span_getter` "
                "parameter and the `labels` parameter."
            )

        super().__init__(
            nlp=nlp,
            name=name,
            span_setter=span_setter or {},
        )
        self.infer_span_setter = (
            span_setter is None if infer_span_setter is None else infer_span_setter
        )

        self.embedding = embedding
        self.labels = labels
        self.linear = torch.nn.Linear(
            self.embedding.output_size,
            0 if labels is None else (len(labels) * 5),
        )
        self.crf = MultiLabelBIOULDecoder(
            1,
            with_start_end_transitions=window < 1,
            learnable_transitions=False,
        )
        self.mode = mode
        if stride is None:
            stride = window // 2
        stride = stride if window > 0 else 0
        if window < stride:
            raise ValueError(
                "The window size must be greater than or equal to the stride."
            )
        if window == 1 and self.mode != "independent":
            warnings.warn(
                "The TrainableNerCrf module will be using a window size equals to 1"
                "(i.e. assumes tags are independent) while trained in non "
                "`independent` mode. This may lead to degraded performance."
            )
        self.window: int = window
        self.stride: int = stride

        self.target_span_getter: Union[
            SpanGetterMapping,
            Callable[[Doc], Iterable[Span]],
        ] = target_span_getter

    def post_init(self, docs: Iterable[Doc], exclude: Set[str]):
        """
        Update the labels based on the data and the span getter,
        and fills in the to_ents and to_span_groups if necessary

        Parameters
        ----------
        docs: Iterable[Doc]
            The documents to use to infer the labels
        exclude: Set[str]
            Components to exclude from the post initialization
        """
        super().post_init(docs, exclude)

        if self.labels is not None and not self.infer_span_setter:
            return

        inferred_labels = set()

        span_setter = defaultdict(lambda: [])

        for doc in docs:
            if callable(self.target_span_getter):
                for ent in self.target_span_getter(doc):
                    inferred_labels.add(ent.label_)
            else:
                for key, span_filter in self.target_span_getter.items():
                    candidates = doc.spans.get(key, ()) if key != "ents" else doc.ents
                    filtered_candidates = (
                        candidates
                        if span_filter is True
                        else [
                            span
                            for span in candidates
                            if span.label_ in span_filter
                            and (self.labels is None or span.label_ in self.labels)
                        ]
                    )
                    for span in filtered_candidates:
                        inferred_labels.add(span.label_)
                        span_setter[key].append(span.label_)
        if self.labels is not None:
            if inferred_labels != set(self.labels):
                warnings.warn(
                    "The labels inferred from the data are different from the "
                    "labels passed to the component. Differing labels are "
                    f"{sorted(set(self.labels) ^ inferred_labels)}",
                    UserWarning,
                )
        else:
            self.update_labels(sorted(inferred_labels))

        self.span_setter = (
            {
                **self.span_setter,
                **{
                    key: sorted(set(value))
                    for key, value in span_setter.items()
                    if key not in self.span_setter
                },
            }
            if self.infer_span_setter
            else self.span_setter
        )

        if not self.labels:
            raise ValueError(
                "No labels were inferred from the data. Please check your data and "
                f"the `target_span_getter` parameter ({self.target_span_getter})."
            )

    def update_labels(self, labels: Sequence[str]):
        old_labels = self.labels if self.labels is not None else ()
        n_old = len(old_labels)
        dict(
            reversed(
                (
                    *zip(old_labels, range(n_old)),
                    *zip(labels, range(n_old, n_old + len(labels))),
                )
            )
        )
        old_index = (
            torch.arange(len(old_labels) * 5)
            .view(-1, 5)[
                [labels.index(label) for label in old_labels if label in labels]
            ]
            .view(-1)
        )
        new_index = (
            torch.arange(len(labels) * 5)
            .view(-1, 5)[
                [old_labels.index(label) for label in old_labels if label in labels]
            ]
            .view(-1)
        )
        new_linear = torch.nn.Linear(self.embedding.output_size, len(labels) * 5)
        new_linear.weight.data[new_index] = self.linear.weight.data[old_index]
        new_linear.bias.data[new_index] = self.linear.bias.data[old_index]
        self.linear.weight = new_linear.weight
        self.linear.bias = new_linear.bias

        # Update initialization arguments
        self.labels = labels

    @property
    def cfg(self):
        return {
            "labels": self.labels,
            "span_setter": self.span_setter,
            "infer_span_setter": self.infer_span_setter,
            "mode": self.mode,
            "window": self.window,
            "stride": self.stride,
        }

    def preprocess(self, doc):
        if self.labels is None:
            raise ValueError(
                "The component was not initialized with any labels. Please "
                "initialize it with the `labels` parameter, pass a list of "
                "labels to the `update_labels` method, or call the `post_init` "
                "method with a list of documents."
            )
        return {
            "embedding": self.embedding.preprocess(doc),
            "length": len(doc),
        }

    def preprocess_supervised(self, doc):
        embedded_spans = list(get_spans(doc, self.embedding.span_getter))
        tags = []

        discarded = []
        for embedded_span, target_ents in zip(
            embedded_spans,
            align_spans(
                list(self.get_target_spans(doc)),
                embedded_spans,
            ),
        ):
            span_tags = [[0] * len(self.labels) for _ in range(len(embedded_span))]
            start = embedded_span.start
            by_label = defaultdict(list)
            for ent in target_ents:
                by_label[ent.label_].append(ent)
            filtered = []
            for label, spans in by_label.items():
                filtered[len(filtered) :], discarded[len(discarded) :] = filter_spans(
                    by_label[label], return_discarded=True
                )

            for ent in filtered:
                label_idx = self.labels.index(ent.label_)
                if ent.start == ent.end - 1:
                    span_tags[ent.start - start][label_idx] = 4
                else:
                    span_tags[ent.start - start][label_idx] = 2
                    span_tags[ent.end - 1 - start][label_idx] = 3
                    for i in range(ent.start + 1 - start, ent.end - 1 - start):
                        span_tags[i][label_idx] = 1
            tags.append(span_tags)

        if discarded:
            warnings.warn(
                f"Some spans in were discarded {doc._.note_id} ("
                f"{', '.join(repr(d.text) for d in discarded)}) because they "
                f"were overlapping with other spans with the same label."
            )

        return {
            **self.preprocess(doc),
            "targets": tags,
        }

    def collate(self, preps) -> NERBatchInput:
        collated: NERBatchInput = {
            "embedding": self.embedding.collate(preps["embedding"]),
        }
        max_len = max(preps["length"])
        if "targets" in preps:
            targets = torch.as_tensor(
                [
                    row + [[-1] * len(self.labels)] * (max_len - len(row))
                    for sample_targets in preps["targets"]
                    for row in sample_targets
                ],
                dtype=torch.long,
            )
            # targets = (targets.unsqueeze(-1) == torch.arange(5)).to(device)
            # mask = (targets[:, 0] != -1).to(device)
            collated["targets"] = targets
        else:
            if self.window > 1:
                win_indices, win_indexer = make_windows(
                    preps["length"],
                    self.window,
                    self.stride,
                )
                collated["window_indices"] = win_indices
                collated["window_indexer"] = win_indexer
        return collated

    def forward(self, batch: NERBatchInput) -> NERBatchOutput:
        encoded = self.embedding.module_forward(batch["embedding"])
        embeddings = encoded["embeddings"]
        mask = encoded["mask"]
        # batch words (labels tags) -> batch words labels tags
        num_samples, num_words = embeddings.shape[:-1]
        num_labels = len(self.labels)
        scores = self.linear(embeddings).view((num_samples, num_words, num_labels, 5))
        loss = tags = None
        if "targets" in batch:
            if self.mode == "independent":
                loss = torch.nn.functional.cross_entropy(
                    scores.view(-1, 5),
                    batch["targets"].view(-1),
                    ignore_index=-1,
                    reduction="sum",
                )
            elif self.mode == "joint":
                loss = self.crf(
                    scores,
                    mask,
                    batch["targets"].unsqueeze(-1) == torch.arange(5).to(scores.device),
                ).sum()
            elif self.mode == "marginal":
                loss = torch.nn.functional.cross_entropy(
                    self.crf.marginal(
                        scores,
                        mask,
                    ).view(-1, 5),
                    batch["targets"].view(-1),
                    ignore_index=-1,
                    reduction="sum",
                )
        else:
            if self.window == 1:
                tags = scores.argmax(-1).masked_fill(~mask.unsqueeze(-1), 0)
            elif self.window <= 0:
                tags = self.crf.decode(scores, mask)
            else:
                win_scores = scores.view(num_samples * num_words, num_labels, 5)[
                    batch["window_indices"]
                ]
                win_tags = self.crf.decode(win_scores, batch["window_indices"] != -1)
                tags = win_tags.view(win_tags.shape[0] * win_tags.shape[1], num_labels)[
                    batch["window_indexer"]
                ]
                tags = tags.view(num_samples, num_words, num_labels)
                tags = tags.masked_fill(~mask.unsqueeze(-1), 0)

            # tags = scores.argmax(-1).masked_fill(~mask.unsqueeze(-1), 0)
        if loss is not None and loss.item() > 100000:
            warnings.warn("The loss is very high, this is likely a tag encoding issue.")
        return {
            "loss": loss,
            "tags": tags,
        }

    def get_target_spans(self, doc) -> Iterable[Span]:
        return (
            self.target_span_getter(doc)
            if callable(self.target_span_getter)
            else get_spans(doc, self.target_span_getter)
        )

    def postprocess(self, docs: List[Doc], batch: NERBatchOutput):
        spans: Dict[Doc, list[Span]] = defaultdict(list)
        embedded_spans = [
            span
            for doc in docs
            for span in list(get_spans(doc, self.embedding.span_getter))
        ]
        for embedded_span_idx, label_idx, start, end in self.crf.tags_to_spans(
            batch["tags"].cpu()
        ).tolist():
            span = embedded_spans[embedded_span_idx][start:end]
            span.label_ = self.labels[label_idx]
            spans[span.doc].append(span)
        for doc in docs:
            self.set_spans(doc, spans.get(doc, []))
        return docs



import shutil
from typing import (
    Optional,
    Sequence,
    Union,
)

import pytest
import spacy.tokenizer
import torch.nn
from confit import Config
from confit.utils.random import set_seed
from spacy.tokens import Span

from edsnlp.core.registries import registry
from edsnlp.data.converters import AttributesMappingArg, get_current_tokenizer
from edsnlp.optimization import LinearSchedule, ScheduledOptimizer
from edsnlp.train import GenericScorer, Reader, train
from edsnlp.utils.span_getters import SpanSetterArg, set_spans


@registry.factory.register("myproject.custom_dict2doc", spacy_compatible=False)
class PseudoDictReader:
    def __init__(
        self,
        *,
        tokenizer: Optional[spacy.tokenizer.Tokenizer] = None,
        name: str = "eds_pseudo.read_dict",
        span_setter: SpanSetterArg = {"ents": True, "*": True},
        bool_attributes: Union[str, Sequence[str]] = [],
        span_attributes: Optional[AttributesMappingArg] = None,
    ):
        self.tokenizer = tokenizer
        self.name = name
        self.span_setter = span_setter
        self.bool_attributes = bool_attributes
        self.span_attributes = span_attributes

    def __call__(self, obj):
        tok = get_current_tokenizer() if self.tokenizer is None else self.tokenizer
        doc = tok(obj["note_text"] or "")
        doc._.note_id = obj.get("note_id", obj.get("__FILENAME__"))
        doc._.note_datetime = obj.get("note_datetime")

        spans = []

        if self.span_attributes is not None:
            for dst in self.span_attributes.values():
                if not Span.has_extension(dst):
                    Span.set_extension(dst, default=None)

        for ent in obj.get("entities") or ():
            ent = dict(ent)
            span = doc.char_span(
                ent.pop("start"),
                ent.pop("end"),
                label=ent.pop("label"),
                alignment_mode="expand",
            )
            for label, value in ent.items():
                new_name = (
                    self.span_attributes.get(label, None)
                    if self.span_attributes is not None
                    else label
                )
                if self.span_attributes is None and not Span.has_extension(new_name):
                    Span.set_extension(new_name, default=None)

                if new_name:
                    span._.set(new_name, value)
            spans.append(span)

        set_spans(doc, spans, span_setter=self.span_setter)
        for attr in self.bool_attributes:
            for span in spans:
                if span._.get(attr) is None:
                    span._.set(attr, False)
        return doc


def test_ner_qualif_train(run_in_test_dir, tmp_path):
    set_seed(42)
    config = Config.from_disk("ner_qlf_config.cfg")
    shutil.rmtree(tmp_path, ignore_errors=True)
    kwargs = config["train"].resolve(registry=registry, root=config)
    nlp = train(**kwargs, output_path=tmp_path, cpu=True)
    scorer = GenericScorer(**kwargs["scorer"])
    last_scores = scorer(nlp, Reader(**kwargs["val_data"])(nlp))

    assert last_scores["exact_ner"]["micro"]["f"] > 0.5
    assert last_scores["qualifier"]["micro"]["f"] > 0.5


def test_qualif_train(run_in_test_dir, tmp_path):
    set_seed(42)
    config = Config.from_disk("qlf_config.cfg")
    shutil.rmtree(tmp_path, ignore_errors=True)
    kwargs = config["train"].resolve(registry=registry, root=config)
    nlp = train(**kwargs, output_path=tmp_path, cpu=True)
    scorer = GenericScorer(**kwargs["scorer"])
    last_scores = scorer(nlp, Reader(**kwargs["val_data"])(nlp))

    assert last_scores["qualifier"]["micro"]["f"] > 0.5


def test_optimizer():
    net = torch.nn.Linear(10, 10)
    optim = ScheduledOptimizer(
        torch.optim.AdamW(
            [
                {
                    "params": list(net.parameters()),
                    "lr": 9e-4,
                    "schedules": LinearSchedule(
                        total_steps=10,
                        warmup_rate=0.1,
                        start_value=0,
                    ),
                }
            ]
        )
    )
    for param in net.parameters():
        assert "exp_avg" not in optim.optim.state[param]
    optim.initialize()
    for param in net.parameters():
        assert "exp_avg" in optim.optim.state[param]
    lr_values = [optim.optim.param_groups[0]["lr"]]
    for i in range(10):
        optim.step()
        lr_values.append(optim.optim.param_groups[0]["lr"])

    # close enough
    assert lr_values == pytest.approx(
        [
            0.0,
            0.0009,
            0.0008,
            0.0007,
            0.0006,
            0.0005,
            0.0004,
            0.0003,
            0.0002,
            0.0001,
            0.0,
        ]
    )



from __future__ import annotations

import logging
import os
import pickle
import warnings
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

import foldedtensor as ft
import torch
import torch.nn.functional as F
from spacy.tokens import Doc, Span
from typing_extensions import NotRequired, TypedDict

from edsnlp.core import PipelineProtocol
from edsnlp.core.torch_component import BatchInput, BatchOutput, TorchComponent
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    SpanEmbeddingComponent,
)
from edsnlp.utils.bindings import (
    BINDING_GETTERS,
    BINDING_SETTERS,
    Qualifiers,
    QualifiersArg,
)
from edsnlp.utils.span_getters import SpanFilter, get_spans

logger = logging.getLogger(__name__)

SpanQualifierBatchInput = TypedDict(
    "SpanQualifierBatchInput",
    {
        "embedding": BatchInput,
        "targets": NotRequired[torch.Tensor],
    },
)
"""
embeds: torch.FloatTensor
    Token embeddings to predict the tags from
mask: torch.BoolTensor
    Mask of the sequences
spans: torch.LongTensor
    2d tensor of n_spans * (doc_idx, ner_label_idx, begin, end)
targets: NotRequired[List[torch.LongTensor]]
    list of 2d tensor of n_spans * n_combinations (1 hot)
"""


class TrainableSpanQualifier(
    TorchComponent[BatchOutput, SpanQualifierBatchInput], BaseComponent
):
    """
    The `eds.span_qualifier` component is a trainable qualifier predictor. In EDS-NLP,
    we call span attributes "qualifiers". In this context, the span qualification task
    consists in assigning values (boolean, strings or any complex object) to
    attributes/extensions of spans such as:

    - `span._.negation`,
    - `span._.date.mode`
    - `span._.cui`

    In the rest of this page, we will refer to a pair of (qualifier, value) as a
    "binding". For instance, the binding `("_.negation", True)` means that the
    qualifier `negation` of the span is (or should be, when predicted) set to `True`.

    Architecture
    ------------
    The model performs span classification by:

    1. Calling a word pooling embedding such as `eds.span_pooler` to compute a single
    embedding for each span
    2. Computing logits for each possible binding using a linear layer
    3. Splitting these bindings into groups of exclusive values such as

        - `event=start` and `event=stop`
        - `negated=False` and `negated=True`

        Note that the above groups are not exclusive, but the values within each group
        are.

    4. Applying the best scoring binding in each group to each span

    Examples
    --------
    To create a span qualifier component, you can use the following code:

    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.span_qualifier(
            # To embed the spans, we will use a span pooler
            embedding=eds.span_pooler(
                pooling_mode="mean",  # mean pooling
                span_getter=["ents", "sc"],
                # that will use a transformer to embed the doc words
                embedding=eds.transformer(
                    model="prajjwal1/bert-tiny",
                    window=128,
                    stride=96,
                ),
            ),
            # For every span embedded by the span pooler
            # (doc.ents and doc.spans["sc"]), we will predict both
            # span._.negation and span._.event_type
            qualifiers=["_.negation", "_.event_type"],
        ),
        name="qualifier",
    )
    ```

    To infer the values of the qualifiers, you can use the pipeline `post_init` method:

    ```{ .python .no-check }
    nlp.post_init(gold_data)
    ```

    To train the model, refer to the [Training](/tutorials/training.md) tutorial.

    You can inspect the bindings that will be used for training and prediction
    ```{ .python .no-check }
    print(nlp.pipes.qualifier.bindings)
    # list of (qualifier name, span labels or True if all, values)
    # Out: [
    #   ('_.negation', True, [True, False]),
    #   ('_.event_type', True, ['start', 'stop'])
    # ]
    ```

    You can also change these values and update the bindings by calling the
    `update_bindings` method. Don't forget to retrain the model if new values are
    added !

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object
    name : str
        Name of the component
    embedding : SpanEmbeddingComponent
        The word embedding component
    qualifiers : QualifiersArg
        The qualifiers to predict or train on. If a dict is given, keys are the
        qualifiers and values are the labels for which the qualifier is allowed, or True
        if the qualifier is allowed for all labels.
    keep_none : bool
        If False, skip spans for which a qualifier returns None. If True (default), the
        None values will be learned and predicted, just as any other value.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: str = "span_qualifier",
        *,
        embedding: SpanEmbeddingComponent,
        qualifiers: QualifiersArg,
        values: Optional[Dict[str, List[Any]]] = None,
        keep_none: bool = False,
    ):
        qualifiers: Qualifiers
        self.values = values
        self.keep_none = keep_none
        self.bindings: List[Tuple[str, List[str], List[Any]]] = [
            (k if k.startswith("_.") else f"_.{k}", v, [])
            for k, v in qualifiers.items()
        ]

        super().__init__(nlp, name)
        self.embedding = embedding
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            self.classifier = torch.nn.Linear(embedding.output_size, 0)

    @property
    def qualifiers(self) -> Qualifiers:
        return {qlf: labels for qlf, labels, _ in self.bindings}

    @qualifiers.setter
    def qualifiers(self, value: Qualifiers):
        bindings = []
        for qlf, labels in value.items():
            groups = [group for group in self.bindings if group[0] == qlf]
            if len(groups) > 1:
                raise ValueError(
                    f"Qualifier {qlf} has different label filters: "
                    f"{[g[0] for g in groups]}. Please use the `update_bindings` "
                    f"method to update the labels."
                )
            if groups:
                bindings.append((qlf, labels, groups[0][2]))
        self.update_bindings(bindings)

    @property
    def span_getter(self):
        return self.embedding.span_getter

    def to_disk(self, path, *, exclude=set()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        # We save the bindings as a pickle file since values can be arbitrary objects
        os.makedirs(path, exist_ok=True)
        data_path = path / "bindings.pkl"
        with open(data_path, "wb") as f:
            pickle.dump(
                {
                    "bindings": self.bindings,
                },
                f,
            )
        return super().to_disk(path, exclude=exclude)

    def from_disk(self, path, exclude=tuple()):
        repr_id = object.__repr__(self)
        if repr_id in exclude:
            return
        # This will receive the directory path + /my_component
        data_path = path / "bindings.pkl"
        with open(data_path, "rb") as f:
            data = pickle.load(f)
        self.update_bindings(data["bindings"])
        self.set_extensions()
        super().from_disk(path, exclude=exclude)

    @property
    def bindings_to_idx(self) -> List[Tuple[str, List[str], Dict[Any, int]]]:
        if getattr(self, "_bindings_to_idx", None) is not None:
            return self._bindings_to_idx
        self._bindings_to_idx = [  # noqa
            (qualifier, labels, {value: idx for idx, value in enumerate(values)})
            for qualifier, labels, values in self.bindings
        ]
        return self._bindings_to_idx

    @property
    def bindings_indexers(self) -> List[Union[Sequence[int], slice]]:
        return self._bindings_indexers

    def set_extensions(self):
        super().set_extensions()
        for group in self.bindings:
            qlf = group[0]
            if qlf.startswith("_."):
                qlf = qlf[2:]
            if not Span.has_extension(qlf):
                Span.set_extension(qlf, default=None)

    def post_init(self, gold_data: Iterable[Doc], exclude: Set[str]):
        super().post_init(gold_data, exclude=exclude)

        bindings = [
            (qlf, labels, dict.fromkeys(vals)) for qlf, labels, vals in self.bindings
        ]
        for doc in gold_data:
            spans = list(get_spans(doc, self.embedding.span_getter))
            for span in spans:
                for qualifier, labels, values in bindings:
                    if labels is True or span.label_ in labels:
                        value = BINDING_GETTERS[qualifier](span)
                        if value is not None or self.keep_none:
                            values[value] = None

        bindings = [
            (qualifier, labels, sorted(values, key=str))
            for qualifier, labels, values in bindings
        ]

        for qualifier, labels, values in bindings:
            if len(values) < 2:
                warnings.warn(
                    f"Qualifier {qualifier} for labels {labels} should have at "
                    f"least 2 values but found {len(values)}: {values}."
                )

        self.update_bindings(bindings)

    def update_bindings(self, bindings: List[Tuple[str, SpanFilter, List[Any]]]):
        keep_bindings = [
            (labels is True or len(labels) > 0) and len(values) > 0
            for k, labels, values in bindings
        ]
        if not all(keep_bindings):
            logger.warning(
                "Some qualifiers have no labels or values and have been removed:"
                + "".join(
                    "\n- " + str(b[0]) + " for labels " + str(b[1])
                    for b, keep in zip(bindings, keep_bindings)
                    if not keep
                )
            )
        bindings = [b for b, keep in zip(bindings, keep_bindings) if keep]
        new_bindings = list(
            dict.fromkeys((k, v) for k, _, group in bindings for v in group)
        )
        old_bindings = list(
            dict.fromkeys((k, v) for k, _, group in self.bindings for v in group)
        )

        new_bindings_to_idx = {binding: idx for idx, binding in enumerate(new_bindings)}
        old_bindings_to_idx = {binding: idx for idx, binding in enumerate(old_bindings)}

        common = [b for b in new_bindings if b in old_bindings]
        old_index = [old_bindings_to_idx[label] for label in common]
        new_index = [new_bindings_to_idx[label] for label in common]
        new_linear = torch.nn.Linear(self.classifier.in_features, len(new_bindings))
        new_linear.weight.data[new_index] = self.classifier.weight.data[old_index]
        self.classifier.weight = new_linear.weight
        self.classifier.out_features = new_bindings
        missing_bindings = set(new_bindings) - set(old_bindings)
        if missing_bindings and len(old_bindings) > 0:
            warnings.warn(
                f"Added {len(missing_bindings)} new bindings. Consider retraining "
                f"the model to learn these new bindings."
            )

        if hasattr(self.classifier, "bias"):
            new_linear.bias.data[new_index] = self.classifier.bias.data[old_index]
            self.classifier.bias = new_linear.bias

        def simplify_indexer(indexer):
            return (
                slice(indexer[0], indexer[-1] + 1)
                if len(indexer) and list(range(indexer[0], indexer[-1] + 1)) == indexer
                else indexer
            )

        self.bindings = bindings
        self._bindings_indexers = [
            simplify_indexer([new_bindings_to_idx[(qlf, value)] for value in values])
            for qlf, labels, values in bindings
        ]
        self._bindings_to_idx = None

    def preprocess(self, doc: Doc) -> Dict[str, Any]:
        return {
            "embedding": self.embedding.preprocess(doc),
        }

    def preprocess_supervised(self, doc: Doc) -> Dict[str, Any]:
        preps = self.preprocess(doc)
        spans = preps["embedding"]["$spans"]
        return {
            **preps,
            "targets": [
                [
                    values_to_idx.get(BINDING_GETTERS[qlf](span), -100)
                    if labels is True or span.label_ in labels
                    else -100
                    for qlf, labels, values_to_idx in self.bindings_to_idx
                ]
                for span in spans
            ],
        }

    def collate(self, batch: Dict[str, Sequence[Any]]) -> SpanQualifierBatchInput:
        collated: SpanQualifierBatchInput = {
            "embedding": self.embedding.collate(batch["embedding"]),
        }
        if "targets" in batch:
            targets = ft.as_folded_tensor(
                batch["targets"],
                dtype=torch.long,
                full_names=("sample", "span", "group"),
                data_dims=("span", "group"),
            ).as_tensor()
            collated["targets"] = targets.view(len(targets), len(self.bindings))
        return collated

    # noinspection SpellCheckingInspection
    def forward(self, batch: SpanQualifierBatchInput) -> BatchOutput:
        """
        Apply the span classifier module to the document embeddings and given spans to:
        - compute the loss
        - and/or predict the labels of spans

        Parameters
        ----------
        batch: SpanQualifierBatchInput
            The input batch

        Returns
        -------
        BatchOutput
        """
        embedding = self.embedding.module_forward(batch["embedding"])
        span_embeds = embedding["embeddings"]

        binding_scores = self.classifier(span_embeds)

        if "targets" in batch:
            losses = []
            pred = None
        else:
            pred = []
            losses = None

        # For each group, for instance:
        # - `event=start` and `event=stop`
        # - `negated=False` and `negated=True`
        for group_idx, bindings_indexer in enumerate(self.bindings_indexers):
            if "targets" in batch:
                losses.append(
                    F.cross_entropy(
                        binding_scores[:, bindings_indexer],
                        batch["targets"][:, group_idx],
                        reduction="sum",
                    )
                )
                assert not torch.isnan(losses[-1]).any(), "NaN loss"
            else:
                pred.append(binding_scores[:, bindings_indexer].argmax(dim=1))

        return {
            "loss": sum(losses) if losses is not None else None,
            "labels": pred,
        }

    def postprocess(self, docs: Sequence[Doc], batch: BatchOutput) -> Sequence[Doc]:
        # Preprocessed docs should still be in the cache
        spans = [s for doc in docs for s in self.preprocess(doc)["embedding"]["$spans"]]
        # For each prediction group (exclusive bindings)...
        for val_indices, (qlf, labels, values) in zip(batch["labels"], self.bindings):
            # For each span...
            for span, idx in zip(spans, val_indices.tolist()):
                # If the span is not filtered out...
                if labels is True or span.label_ in labels:
                    # ...assign the predicted value to the span
                    BINDING_SETTERS[qlf](span, values[idx])
        return docs



"""`eds.liver_disease` pipeline"""
from typing import Any, Dict, List, Optional, Union

from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg

from ..base import DisorderMatcher
from .patterns import default_patterns


class LiverDiseaseMatcher(DisorderMatcher):
    """
    The `eds.liver_disease` pipeline component extracts mentions of liver disease.

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/liver_disease/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to either
        - `"MILD"` for mild liver diseases
        - `"MODERATE_TO_SEVERE"` else

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.liver_disease())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/liver-disease-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: Union[Dict[str, Any], List[Dict[str, Any]]]
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.liver_disease` component was developed by AP-HP's Data Science team with a
    team of medical experts. A paper describing in details the development of those
    components is being drafted and will soon be available.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "liver_disease",
        *,
        patterns: Union[Dict[str, Any], List[Dict[str, Any]]] = default_patterns,
        label: str = "liver_disease",
        span_setter: SpanSetterArg = {"ents": True, "liver_disease": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            patterns=patterns,
            detailed_status_mapping={
                0: "ABSENT",
                1: "MILD",
                2: "MODERATE_TO_SEVERE",
            },
            label=label,
            span_setter=span_setter,
        )

    def process(self, doc: Doc) -> List[Span]:
        for span in super().process(doc):
            if span._.source in {"moderate_severe", "transplant"}:
                span._.status = 2

            yield span



from typing import List, Optional, Tuple

from spacy.tokens import Doc

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import BaseComponent

from .patterns import quotes_and_apostrophes


class QuotesConverter(BaseComponent):
    """
    We normalise quotes, following this
    `source <https://www.cl.cam.ac.uk/~mgk25/ucs/quotes.html>`_.

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline object.
    name : Optional[str]
        The component name.
    quotes : List[Tuple[str, str]]
        List of quotation characters and their transcription.
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: Optional[str] = "spaces",
        *,
        quotes: List[Tuple[str, str]] = quotes_and_apostrophes,
    ):
        super().__init__(nlp=nlp, name=name)
        self.translation_table = str.maketrans(
            "".join(quote_group for quote_group, _ in quotes),
            "".join(rep * len(quote_group) for quote_group, rep in quotes),
        )

    def __call__(self, doc: Doc) -> Doc:
        """
        Normalises quotes.

        Parameters
        ----------
        doc : Doc
            Document to process.

        Returns
        -------
        Doc
            Same document, with quotes normalised.
        """

        for token in doc:
            token.norm_ = token.norm_.translate(self.translation_table)

        return doc



from __future__ import annotations

import sys
import warnings
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import pandas as pd
from confit import VisibleDeprecationWarning
from spacy.tokens import Doc

import edsnlp.data

from ..core import PipelineProtocol
from ..utils.span_getters import SpanGetterArg, validate_span_setter
from ..utils.spark_dtypes import PySparkPrettyPrinter

if TYPE_CHECKING:
    from databricks.koalas.frame import DataFrame as KoalasDataFrame
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame

ExtensionSchema = Union[
    str,
    List[str],
    Dict[str, Any],
]


def slugify(chained_attr: str) -> str:
    """
    Slugify a chained attribute name

    Parameters
    ----------
    chained_attr : str
        The string to slugify (replace dots by _)

    Returns
    -------
    str
        The slugified string
    """
    return chained_attr.replace(".", "_")


def pipe(
    df: Union[SparkDataFrame, KoalasDataFrame, pd.DataFrame],
    nlp: PipelineProtocol,
    n_jobs: int = -2,
    # Legacy parameters
    context: List[str] = [],
    results_extractor: Optional[Callable[[Doc], List[Dict[str, Any]]]] = None,
    additional_spans: SpanGetterArg = [],
    extensions: ExtensionSchema = [],
    dtypes: Any = None,
    **kwargs,
) -> Union[SparkDataFrame, KoalasDataFrame, pd.DataFrame]:
    """
    Helper to process a pandas, koalas or spark dataframe. This function is deprecated.
    Prefer using the following instead:

    ```{ .python .no-check }
    import edsnlp

    docs = edsnlp.data.from_***(
        df,
        converter='omop',
        doc_attributes=context,
    )
    docs = docs.map_pipeline(nlp)
    res = edsnlp.data.to_***(
        docs,
        converter='ents',  # or custom extractor
        span_getter="ents",
        span_attributes=span_attributes,
        **kwargs
    )
    ```

    You can also call this function to get a migration suggestion.

    Parameters
    ----------
    df: Union[SparkDataFrame, KoalasDataFrame, pd.DataFrame]
        The dataframe to process, can be a pandas, spark or koalas dataframe
    nlp: PipelineProtocol
        The pipeline to use
    n_jobs: int
        Number of CPU workers to use
    context: List[str]
        List of context attributes to keep
    results_extractor: Optional[Callable[[Doc], List[Dict[str, Any]]]]
        Function to extract results from the pipeline. Defaults to one row per
        entities.
    additional_spans: SpanGetterArg
        Additional spans groups to keep, defaults to `ents` (doc.ents)
    extensions: ExtensionSchema
        Span extensions to export as a column. Can be a list of extension names, a dict
        of extension names to types, or a string
    dtypes: Any
        Spark schema to use for the output dataframe. This is only used if
        the input dataframe is a spark dataframe.
    kwargs: Any
        Additional keyword arguments to pass to the `edsnlp.data.to_*` function

    Returns
    -------
    Union[SparkDataFrame, KoalasDataFrame, pd.DataFrame]
        The processed dataframe
    """
    is_pandas = isinstance(df, pd.DataFrame)
    s = (
        "edsnlp.processing.pipe is deprecated, use the following instead:\n\n"
        "import edsnlp\n"
        f"docs = edsnlp.data.from_{'pandas' if is_pandas else 'spark'}(\n"
        f"    df,\n"
        f"    converter='omop',\n"
        f"    doc_attributes={context!r}\n"
        f")\n"
        f"docs = docs.map_pipeline(nlp)\n"
        f"res = edsnlp.data.to_{'pandas' if is_pandas else 'spark'}(\n"
        f"    docs,\n"
        f"    converter={repr('ents') if results_extractor is None else 'extractor'},\n"
    )
    if not results_extractor and additional_spans:
        s += f"    span_getter={repr('ents' if additional_spans is None else [*additional_spans, 'ents'])},\n"  # noqa: E501
    if not results_extractor and extensions:
        s += f"    span_attributes={repr({ext: slugify(ext) for ext in extensions})},\n"
    if dtypes:
        s += f"    dtypes={PySparkPrettyPrinter().pformat(dtypes)},\n"
    s += ")"
    warnings.warn(s, VisibleDeprecationWarning)

    write_span_getter = validate_span_setter(
        "ents" if additional_spans is None else [*additional_spans, "ents"],
    )

    if isinstance(df, pd.DataFrame):
        docs = edsnlp.data.from_pandas(
            df,
            converter="omop",
            doc_attributes=context,
        )
        docs = docs.map_pipeline(nlp)
        docs = docs.set_processing(num_cpu_workers=n_jobs)
        return (
            edsnlp.data.to_pandas(
                docs,
                converter="ents",
                span_getter=write_span_getter,
                span_attributes={ext: slugify(ext) for ext in extensions},
                **kwargs,
            )
            if results_extractor is None
            else edsnlp.data.to_pandas(
                docs,
                converter=results_extractor,
            )
        )
    import pyspark.sql.types as T

    try:
        KoalasDataFrame = sys.modules["databricks.koalas.frame"].DataFrame
    except (AttributeError, KeyError):
        KoalasDataFrame = None
    is_koalas = KoalasDataFrame and isinstance(df, KoalasDataFrame)  # type: ignore
    if is_koalas:
        df: SparkDataFrame = df.to_spark()  # type: ignore

    docs = edsnlp.data.from_spark(
        df,
        converter="omop",
        doc_attributes=context,
    )
    docs = docs.map_pipeline(nlp)

    schema = None

    if results_extractor is None:
        if dtypes is None and (not extensions or isinstance(extensions, dict)):
            schema = [
                df.schema["note_id"],
                T.StructField("lexical_variant", T.StringType(), False),
                T.StructField("label", T.StringType(), False),
                T.StructField("start", T.IntegerType(), False),
                T.StructField("end", T.IntegerType(), False),
                T.StructField("span_type", T.StringType(), False),
            ]

        if isinstance(extensions, dict):
            schema = [
                *schema,
                *(
                    T.StructField(slugify(extension_name), extension_type, True)
                    for extension_name, extension_type in extensions.items()
                ),
            ]
        converter = "ents"
    else:
        if dtypes:
            schema = [
                df.schema["note_id"],
            ] + [
                T.StructField(slugify(extension_name), extension_type, True)
                for extension_name, extension_type in dtypes.items()
            ]

        def converter(doc):
            res = results_extractor(doc)
            return (
                [{"note_id": doc._.note_id, **row} for row in res]
                if isinstance(res, list)
                else {"note_id": doc._.note_id, **res}
            )

    schema = T.StructType(schema) if schema is not None else None

    docs = (
        edsnlp.data.to_spark(
            docs,
            converter="ents",
            span_getter=write_span_getter,
            span_attributes={ext: slugify(ext) for ext in extensions},
            dtypes=schema,
        )
        if results_extractor is None
        else edsnlp.data.to_spark(
            docs,
            converter=converter,
            dtypes=schema,
        )
    )

    if is_koalas:
        docs = docs.to_koalas()

    return docs



from typing import List

import spacy
from fastapi import FastAPI
from pydantic import BaseModel

import edsnlp

app = FastAPI(title="EDS-NLP", version=edsnlp.__version__)

nlp = spacy.blank("eds")

nlp.add_pipe("eds.sentences")

config = dict(
    regex=dict(
        covid=[
            "covid",
            r"covid[-\s]?19",
            r"sars[-\s]?cov[-\s]?2",
            r"corona[-\s]?virus",
        ],
    ),
    attr="LOWER",
)
nlp.add_pipe("eds.matcher", config=config)

nlp.add_pipe("eds.negation")
nlp.add_pipe("eds.family")
nlp.add_pipe("eds.hypothesis")
nlp.add_pipe("eds.reported_speech")


class Entity(BaseModel):  # (2)

    # OMOP-style attributes
    start: int
    end: int
    label: str
    lexical_variant: str
    normalized_variant: str

    # Qualifiers
    negated: bool
    hypothesis: bool
    family: bool
    reported_speech: bool


class Document(BaseModel):  # (1)
    text: str
    ents: List[Entity]


@app.post("/process", response_model=List[Document])  # (1)
async def process(
    notes: List[str],  # (2)
):

    documents = []

    for doc in nlp.pipe(notes):
        entities = []

        for ent in doc.ents:
            entity = Entity(
                start=ent.start_char,
                end=ent.end_char,
                label=ent.label_,
                lexical_variant=ent.text,
                normalized_variant=ent._.normalized_variant,
                negated=ent._.negation,
                hypothesis=ent._.hypothesis,
                family=ent._.family,
                reported_speech=ent._.reported_speech,
            )
            entities.append(entity)

        documents.append(
            Document(
                text=doc.text,
                ents=entities,
            )
        )

    return documents



from edsnlp.core import registry
from edsnlp.pipes.ner.scores.base_score import SimpleScoreMatcher

DEFAULT_CONFIG = dict(
    regex=None,
    attr="NORM",
    value_extract=None,
    score_normalization=None,
    window=7,
    ignore_excluded=False,
    ignore_space_tokens=False,
    flags=0,
    span_setter={"ents": True},
)

create_component = registry.factory.register(
    "eds.score",
    assigns=["doc.ents", "doc.spans"],
    deprecated=["score"],
)(SimpleScoreMatcher)



"""
Process CIM10 patterns.

!!! warning "Watch out for the encoding"

    We had to convert the CIM-10 file from windows-1252 to utf-8.

Source: https://www.atih.sante.fr/plateformes-de-transmission-et-logiciels/logiciels-espace-de-telechargement/id_lot/456
"""  # noqa

from pathlib import Path

import pandas as pd
import typer


def run(
    raw: Path = typer.Argument(..., help="Path to the raw file"),
    output: Path = typer.Option(
        "edsnlp/resources/cim10.csv.gz", help="Path to the output CSV table."
    ),
) -> None:
    """
    Convenience script to automatically process the CIM10 terminology
    into a processable file.
    """

    df = pd.read_csv(raw, sep="|", header=None)

    typer.echo(f"Processing {len(df)} French ICD codes...")

    df.columns = ["code", "type", "ssr", "psy", "short", "long"]
    for column in ["code", "short", "long"]:
        df[column] = df[column].str.strip()

    typer.echo(f"Saving to {output}")

    df.to_csv(output, index=False)

    typer.echo("Done !")


if __name__ == "__main__":
    typer.run(run)
